# -*- mode: python; coding: utf-8 -*-
# vim:smartindent cinwords=if,elif,else,for,while,try,except,finally,def,class:ts=4:sts=4:sta:et:ai:shiftwidth=4
#
# Copyright ©  2004 Canonical Ltd.
#	Author: Robert Collins <robertc@robertcollins.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from copy import copy
import errno
import logging
import os
import re
import stat
import string
import sys

from pqm.core import do_mkdir
from pqm.errors import PQMException, PQMTlaFailure
from pqm.script import Script

__version__ = "0.5"

# These need thought and layer - but for now, are just moved
# from the binary script intact. They should be made class or
# instance scope, but there wasn't a trivial way while moving
# them from the script.
used_transactions = {}
gnupatch_path = 'patch'
logger = logging # default value for simple use.
groups = {}


def get_queuedir(config_parser, logger, args):
    """Get the queuedir that should be used from the config"""
    if config_parser.has_option('DEFAULT', 'queuedir'):
        return os.path.abspath(os.path.expanduser(
            config_parser.get('DEFAULT', 'queuedir')))
    elif len(args) > 0:
        return os.path.abspath(args[0])
    else:
        logger.error("No queuedir specified on command line or"
                     " in config_parser files.")
        sys.exit(1)


def verify_sig(sender, msg, sig, check_replay, logger, keyring):
    """Verify the GPG signature on a message."""
    verifier = GPGSigVerifier([keyring], gpgv=gpgv_path)
    try:
        tmp_msgpath = os.path.join(pqm_subdir,'tmp-msg')
        open(tmp_msgpath, 'w').write(msg)
        if sig:
            tmp_sigpath = os.path.join(pqm_subdir,'tmp-sig')
            open(tmp_sigpath, 'w').write(sig)
        else:
            tmp_sigpath = None
        output = verifier.verify(tmp_msgpath, tmp_sigpath)
        os.unlink(tmp_msgpath)
        if sig:
            os.unlink(tmp_sigpath)
    except GPGSigVerificationFailure, e:
        raise PQMException(sender, "Failed to verify signature: %s" % e._value)
    gpgre = re.compile('^\[GNUPG:\] (SIG_ID.+)$')
    sigid = None
    for line in output:
        match = gpgre.match(line)
        if match:
            sigid = match.group(1)
            break
    if not sigid:        
        raise PQMException(sender, "Couldn't determine signature timestamp")
    if check_replay and used_transactions.has_key(sigid):
        logger.error("Replay attack detected, aborting")
        raise PQMException(sender, "Replay attack detected, aborting")
    gpg_key_re = re.compile('^\[GNUPG:\] GOODSIG ([0-9A-F]+) .*<([^>]*)>.*$')
    for line in output:
        match = gpg_key_re.match(line)
        if match:
            sig_from_mail = match.group (2)
            sig_from_key = match.group (1)
            break
    logger.info ("garh %s : %s", sig_from_key, sig_from_mail)
    return sigid, sig_from_mail

class GPGSigVerificationFailure(Exception):
    """GPG Verification failed."""
    def __init__(self, value, output):
        self._value = value
        self._output = output
        
    def __str__(self):
        return `self._value`

    def getOutput(self):
        return self._output

class GPGSigVerifier:
    """A class to verify GPG signatures."""
    def __init__(self, keyrings, gpgv=None):
        self._keyrings = keyrings
        if gpgv is None:
            gpgv = '/usr/bin/gpgv'
        self._gpgv = gpgv

    def verify(self, filename, sigfilename=None):
        """Verify the signature on tjhe file filename, detached signatures on
        sigfilename.
        """
        (stdin, stdout) = os.pipe()
        pid = os.fork()
        if pid == 0:
            os.close(stdin)
            os.dup2(stdout, 1)
            os.dup2(stdout, 2)
            args = [self._gpgv]
            args.append('--status-fd=2')
            for keyring in self._keyrings:
                args.append('--keyring')
                args.append(keyring)
            if sigfilename:
                args.append(sigfilename)
            args.append(filename)
            os.execvp(self._gpgv, args)
            os.exit(1)
        os.close(stdout)
        output = os.fdopen(stdin).readlines()
        (pid, status) = os.waitpid(pid, 0)
        if not (status is None 
                or (os.WIFEXITED(status) and os.WEXITSTATUS(status) == 0)):
            if os.WIFEXITED(status):
                msg = ("gpgv exited with error code %d" % 
                        os.WEXITSTATUS(status))
            elif os.WIFSTOPPED(status):
                msg = ("gpgv stopped unexpectedly with signal %d" %
                        os.WSTOPSIG(status))
            elif os.WIFSIGNALED(status):
                msg = "gpgv died with signal %d" % os.WTERMSIG(status)
            raise GPGSigVerificationFailure(msg, output)
        return output


def popen_noshell_with_input(cmd, inputfd, *args):
    (stdin, stdout) = os.pipe()
    pid = os.fork()
    if pid == 0:
        try:
            os.close(stdin)
            if inputfd is None:
                inputfd = os.open('/dev/null', os.O_RDONLY) 
            os.dup2(inputfd, 0)
            os.dup2(stdout, 1)
            os.dup2(stdout, 2)
            logger.info("running: " + string.join([cmd] + list(args),' '))
            os.execvp(cmd, [cmd] + list(args))
        except:
            os._exit(1)
    os.close(stdout)
    output = os.fdopen(stdin).readlines()
    (pid, status) = os.waitpid(pid, 0)
    msg = ''
    if not (status is None or (os.WIFEXITED(status) and os.WEXITSTATUS(status) == 0)):
        if os.WIFEXITED(status):
            msg = "%s exited with error code %d" % (cmd, os.WEXITSTATUS(status),)
        elif os.WIFSTOPPED(status):
            msg = "%s stopped unexpectedly with signal %d" % (cmd, os.WSTOPSIG(status),)
        elif os.WIFSIGNALED(status):
            msg = "%s died with signal %d" % (cmd, os.WTERMSIG(status),)
    return (status, msg, output)

def popen_noshell(cmd, *args):
    return apply(popen_noshell_with_input, [cmd, None] + list(args))

class VCSHandler:

    def branch_exists(self, sender, branch):
        raise NotImplementedError("branch_exists must be provided by VCS"
                                  " Handlers.")
        
    def commit(self, sender, dir, summary, branch, config):
        raise PQMTlaFailure(sender, 'Unsupported operation')
        
    def do_star_merge(self, sender, from_branch, local_dir):
        raise PQMTlaFailure(sender, 'Unsupported operation')

    def do_replay(self, sender, from_branch, local_dir):
        raise PQMTlaFailure(sender, 'Unsupported operation')

    def do_repo_cache(self, sender, fromrepo, fromrevision):
        raise PQMTlaFailure(sender, 'Unsupported operation')
    
    def do_repo_uncache(self, sender, fromrepo, fromrevision):
        raise PQMTlaFailure(sender, 'Unsupported operation')

    def do_tag(self, sender, fromrepo, fromrevision, to_repo, to_revision):
        raise PQMTlaFailure(sender, 'Unsupported operation')
    
    def do_create_branch(self, sender, to_repo, to_revision):
        raise PQMTlaFailure(sender, 'Unsupported operation')
    
    def do_make_repo(self, sender, repo, location):
        raise PQMTlaFailure(sender, 'Unsupported operation')
    
    def do_create_version(self, sender, repo, revision):
        raise PQMTlaFailure(sender, 'Unsupported operation')

    def make_local_dir(self, sender, branch, output_dir):
        raise PQMTlaFailure(sender, 'Unsupported operation')

    def latest_revision_info(self, sender, branch):
        raise PQMTlaFailure(sender, 'Unsupported operation')


class ArchHandler(VCSHandler):
     
    def branch_exists(self, sender, branch):
        # XXX: This isn't quite right, as makeing a new branch
        # also calls validate_revision. It really needs two 
        # separate methods
        try:
            # FIXME, check the revision/version exists
            repo, revision = branch.split('/', 1)
            self.do_whereis_repo(sender, repo)
            return True
        except PQMTlaFailure, e:
            self.last_error = e

    def commit(self, sender, dir, summary, branch, config):
        self.run_in_dir(dir, runtla, sender, 'commit', '-s', summary)
        logger.info("commit succeeded")

    def do_replay(self, sender, from_branch, local_dir):
        return self.run_in_dir(local_dir, runtla, sender, 'replay',
                               '%s' % from_branch)

    def do_whereis_repo(self, sender, repo):
        return runtla(sender, 'whereis-archive', '%s' % (repo))
    
    def run_in_dir(self, local_dir, method, *args):
        """call method after chdiring to local_dir, then chdir back."""
        orig_dir = os.getcwdu()
        try:
            os.chdir(local_dir)
            return method(*args)
        finally:
            os.chdir(orig_dir)

    def make_local_dir(self, sender, branch, output_dir):
        runtla(sender, 'get', '%s' % branch, output_dir)


class ArXHandler(ArchHandler):

    def do_star_merge(self, sender, from_branch, local_dir):
        return self.run_in_dir(local_dir, runtla, sender, 'merge',
                               '%s' % from_branch)

    def do_repo_cache(self, sender, fromrepo, fromrevision):
        return runtla(sender, 'archive-cache', '--add',
                      '%s/%s' % (fromrepo, fromrevision))
    
    def do_repo_uncache(self, sender, fromrepo, fromrevision):
        return runtla(sender, 'archive-cache', '--delete',
                      '%s/%s' % (fromrepo, fromrevision))

    def do_tag(self, sender, fromrepo, fromrevision, to_repo, to_revision):
        return runtla(sender, 'tag',
                      '%s/%s' % (to_repo, to_revision),
                      '%s/%s' % (fromrepo, fromrevision))
    
    def do_create_branch(self, sender, to_repo, to_revision):
        return runtla(sender, 'fork',
                      '%s/%s' % (to_repo, to_revision))
    
    def do_make_repo(self, sender, repo, location):
        return runtla(sender, 'make-archive',
                      '%s' % (repo), '%s' % (location))
    
    def do_create_version(self, sender, repo, revision):
        return runtla(sender, 'init',
                      '%s/%s' % (repo, revision))

    def do_whereis_repo(self, sender, repo):
        return runtla(sender, 'archives', '%s/' % (repo))


class TlaHandler(ArchHandler):

    def do_star_merge(self, sender, from_branch, local_dir):
        return self.run_in_dir(local_dir, runtla, sender, 'star-merge',
                               '%s' % from_branch)

    def do_repo_cache(self, sender, fromrepo, fromrevision):
        return runtla(sender, 'cacherev', '%s/%s' % (fromrepo, fromrevision))
    
    def do_repo_uncache(self, sender, fromrepo, fromrevision):
        return runtla(sender, 'uncacherev', '%s/%s' % (fromrepo, fromrevision))

    def do_tag(self, sender, fromrepo, fromrevision, to_repo, to_revision):
        return runtla(sender, 'tag', '%s/%s' % (fromrepo, fromrevision),
                      '%s/%s' % (to_repo, to_revision))
    
    def do_create_branch(self, sender, to_repo, to_revision):
        return runtla(sender, 'archive-setup', '%s/%s' % (to_repo, to_revision))
    
    def do_make_repo(self, sender, repo, location):
        return runtla(sender, 'make-archive', '%s' % (repo), '%s' % (location))
    
    def do_create_version(self, sender, repo, revision):
        return runtla(sender, 'archive-setup', '%s/%s' % (to_repo, to_revision))
 

class BazBaseHandler(ArchHandler):

    def do_repo_cache(self, sender, fromrepo, fromrevision):
        return runtla(sender, 'cacherev', '%s/%s' % (fromrepo, fromrevision))
    
    def do_repo_uncache(self, sender, fromrepo, fromrevision):
        return runtla(sender, 'uncacherev', '%s/%s' % (fromrepo, fromrevision))

    def do_tag(self, sender, fromrepo, fromrevision, to_repo, to_revision):
        return runtla(sender, 'branch', '%s/%s' % (fromrepo, fromrevision),
                      '%s/%s' % (to_repo, to_revision))

    def do_make_repo(self, sender, repo, location):
        return runtla(sender, 'make-archive', '%s' % (repo), '%s' % (location))
    

class Baz1_0Handler(BazBaseHandler):

    def do_star_merge(self, sender, from_branch, local_dir):
        return self.run_in_dir(local_dir, runtla, sender, 'star-merge', '-t',
                               '%s' % from_branch)

    def do_create_branch(self, sender, to_repo, to_revision):
        return runtla(sender, 'archive-setup', '%s/%s' % (to_repo, to_revision))
    
    def do_create_version(self, sender, repo, revision):
        return runtla(sender, 'archive-setup', '%s/%s' % (to_repo, to_revision))


class Baz1_1Handler(BazBaseHandler):

    def do_star_merge(self, sender, from_branch, local_dir):
        return self.run_in_dir(local_dir, runtla, sender, 'merge',
                               '--star-merge', '%s' % from_branch)


class Bazaar2Handler(VCSHandler):

    def __init__(self):
        try:
            import bzrlib
        except ImportError:
            pass
        else:
            import bzrlib.plugin
            bzrlib.plugin.load_plugins()

    def branch_exists(self, sender, branchspec):
        try:
            import bzrlib
        except ImportError:
            return False
        from bzrlib.branch import Branch
        try:
            branch = Branch.open(branchspec)
            return True
        except bzrlib.errors.NotBranchError, e:
            self.last_error = e
            if not hasattr(e, "output"):
                self.last_error.output = []
            return False

    def commit(self, sender, dir, summary, parent_branch, config):
        from bzrlib.branch import Branch
        import bzrlib.plugin
        import bzrlib.errors
        from bzrlib.workingtree import WorkingTree
        tree = WorkingTree.open(dir)
        branch = tree.branch
        tree.commit(summary, verbose=False)
        logger.info("commit succeeded")
        # config-manager currently always branches. TOFIX
        # this equality tests for 'not a checkout'
        if tree.bzrdir.root_transport.base == branch.bzrdir.root_transport.base:
            parent_branch = Branch.open(parent_branch)
            branch.push(parent_branch)
            logger.info("commited '%s' to '%s'", branch, parent_branch)
        if config.get('publish_to', None) != None:
            try:
                logger.info("publishing to '%s'", config['publish_to'])
                public_branch = Branch.open(config['publish_to'])
                branch.push(public_branch)
                logger.info("commited '%s' to '%s'", branch, public_branch)
            except (bzrlib.errors.BzrError, bzrlib.errors.BzrNewError), e:
                # Discard the successfully committed item from the 'master'
                # branch.
                parent_branch.set_revision_history(
                    parent_branch.revision_history()[:-1])
                raise PQMTlaFailure(sender, [str(e).split('\n')])

    def do_star_merge(self, sender, from_branch, local_dir):
        import bzrlib.errors
        from bzrlib.workingtree import WorkingTree
        tree = WorkingTree.open(local_dir)
        tree.lock_write()
        try:
            try:
                return self._do_star_merge(sender, from_branch, local_dir, tree)
            except bzrlib.errors.UnrelatedBranches:
                raise PQMTlaFailure(sender,
                    ['Branches have no common ancestor, and no merge base'
                     ' revision was specified.'])
        finally:
            tree.unlock()

    def _do_star_merge(self, sender, from_branch, local_dir, tree):
        from bzrlib.merge import Merger, Merge3Merger
        from bzrlib.tag import _merge_tags_if_possible
        merger = Merger(tree.branch, this_tree=tree)
        merger.check_basis(True)
        merger.set_other([from_branch, -1])
        merger.set_base([None, None])
        if merger.base_rev_id == merger.other_rev_id:
            raise PQMTlaFailure(sender, ['Nothing to merge.'])
        merger.backup_files = False
        merger.merge_type = Merge3Merger
        merger.set_interesting_files(None)
        merger.show_base = False
        merger.reprocess = False
        conflicts = merger.do_merge()
        merger.set_pending()
        if conflicts:
            error_lines = ['Conflicts during merge: ']
            tree.lock_read()
            try:
                for conflict in tree.conflicts():
                    error_lines.append(str(conflict))
            finally:
                tree.unlock()
            raise PQMTlaFailure(sender, error_lines)
        _merge_tags_if_possible(merger.other_branch, tree.branch)
        return ["merge successful"]
        
    def make_local_dir(self, sender, branch_spec, output_dir):
        from bzrlib.branch import Branch, BranchReferenceFormat
        import bzrlib.bzrdir as bzrdir
        os.mkdir(output_dir)
        branch = Branch.open(branch_spec)
        branch.create_checkout(output_dir, lightweight=True)

    def latest_revision_info(self, sender, branch_location):
        from bzrlib.branch import Branch
        branch = Branch.open(branch_location)
        return branch.revno(), branch.repository.get_revision(branch.last_revision()).message


class BranchSpecOptionHandler:
    """Configuration mechanism for a branch specifications.
    
    This manages the default and per-branch-specifier options from pqm.conf.
    Each stanza in the configuration may affect many nested branches which is
    why this is a BranchSpec OptionHandler not a BranchOptionHandler.
    """

    def __init__(self, configp, queuedir=None):
        self._configp = configp
        # TODO: remove the queuedir from this class.  The only reason
        # it is here is to have the status file.  This should now be
        # moved into the PatchQueueManager class.
        # Tim Penhey 2008-05-29
        if self._configp.has_option('DEFAULT', 'queuedir'):
            self.queuedir = os.path.abspath(os.path.expanduser(
                self._configp.get('DEFAULT', 'queuedir')))
            do_mkdir(self.queuedir, logger)
            do_mkdir(os.path.join(self.queuedir, 'pqm'), logger)
        else:
            self.queuedir = queuedir
        self._specs = {}
        self._optionmap = {}
        self._optionmap['precommit_hook'] = ['str', None]
        self._optionmap['project'] = ['str', None]
        self._optionmap['published_at'] = ['str', None]
        self._optionmap['publish_to'] = ['str', None]
        self._optionmap['build_config'] = ['str', None]
        self._optionmap['build_dir'] = ['str', None]
        self._optionmap['commiters'] = ['str', None]
        self._optionmap['commit_re'] = ['str', None]

        self.get_groups()

        # find the list of branch specs 
        # by removing group sections and location overrides from the 
        # list of sections.
        sects = self._configp.sections()
        if len(sects) > 0:
            for sect in sects:
                if str(sect) in groups.keys():
                    continue
                if str(sect) == 'location overrides':
                    continue
                self._specs[sect] = {}
        for spec in self._specs:
            self._specs[spec] = self.get_option_map(spec)
        
        if self.queuedir:
            self.status = StatusFile(self.queuedir + '/pqm/status')

    def get_groups(self):
        if self._configp.has_option('DEFAULT', 'groups'):
            for group in self._configp.get('DEFAULT', 'groups').split(','):
                groups[group.strip()]=[]
            logger.info('found groups %s', groups)
            for group in groups.keys():
                for member in self._configp.get(group, 'members').split(','):
                    groups[group].append(member.strip())
                logger.info('group %s has members %s', group, groups[group])

    def get_option_map(self, dist):
        ret = self._specs[dist]
        for key in self._optionmap.keys():
            type = self._optionmap[key][0]
            ret[key] = self._optionmap[key][1]
            if self._configp.has_option ('DEFAULT', key):
                ret[key] = self.get_option (type, 'DEFAULT', key)
            if self._configp.has_option (dist, key):
                ret[key] = self.get_option (type, dist, key)
        return ret

    def get_option(self, type, dist, key):
        if type == 'int':
            return self._configp.getint(dist, key)
        elif type == 'str':
            return self._configp.get(dist, key)
        elif type == 'bool':
            return self._configp.getboolean(dist, key)

        assert(None)

    def _branchspec_matches(self, branch_spec, url):
        return ((branch_spec == url and url[-1] != '/') or 
                (branch_spec.startswith(url) and url[-1] == '/' and 
                 len(branch_spec) > len(url)))

    def get_target_config(self, branch_spec):
        """Get the branch url we should use and its config."""
        for key, value in self._specs.items():
            if (self._branchspec_matches(branch_spec, key) or
                (value.has_key('published_at') and
                value['published_at'] is not None and
                self._branchspec_matches(branch_spec, value['published_at']))):
                if not branch_spec.startswith(key):
                    branch_spec = key + branch_spec[len(value['published_at']):]
                config = copy(value)
                if value.get('published_at', None) is not None:
                    # calculate real published location
                    config['published_at'] = (value['published_at'] + 
                                              branch_spec[len(key):])
                    if config.get('publish_to', None) is None:
                        config['publish_to'] = config['published_at']
                    else:
                        config['publish_to'] = (config['publish_to'] +
                                                branch_spec[len(key):])
                if not config.has_key('publish_to'):
                    config['publish_to'] = None
                return branch_spec, config
        raise KeyError("branch %s is not configured for pqm." % branch_spec)

    def project_names(self):
        """Return a set of the project names present in the configuration."""
        result = set()
        if self._configp.has_option('DEFAULT', 'project'):
            result.add(self._configp.get('DEFAULT', 'project'))
        for spec in self._specs:
            if self._configp.has_option(spec, 'project'):
                result.add(self._configp.get(spec, 'project'))
        return result


class StatusFile(object):
    """A 'status' file. This stores up to 12 lines of info about PQM's status."""

    def __init__(self, filename, height=12):
        """Construct a Status file at the path filename."""
        statusfile = file(filename, 'at+')
        statusfile.seek(0)
        self.height = height
        self.status_area = 0
        self.lines = statusfile.readlines()
        self.lines = self.lines[-self.height:]
        self.statusfile = None
        self.filename = filename

    def line(self, line):
        """Add line to the status file.

        This isn't quite appropriate for here, but there is no better place for
        it to live yet.
        """
        if not line.endswith('\n'):
            line += '\n'
        if self.status_area:
            self.lines.insert(-self.status_area, line)
        else:
            self.lines.append(line)
        self.lines = self.lines[-self.height:]
        self.flush()

    def flush(self):
        """Push self.lines out to disk."""
        if self.statusfile is None:
            # reopen so that seeking works
            self.statusfile = file(self.filename, 'wt')
        self.statusfile.seek(0)
        for line in self.lines:
            self.statusfile.write(line)
        self.statusfile.flush()
        self.statusfile.truncate()

    def set_status_area_height(self, height):
        """Set the numbe of lines at the bottom of the status file for status.

        The status area scrolls independently of the rest of the StatusFile
        lines and doesn't alter the status area. This allows the status area
        to be a summary area that stays constant.

        Use status_line() to add lines to the status area.
        """
        self.status_area = height
        normal_area = self.height - self.status_area
        self.lines = ['\n'] * normal_area + self.lines
        if len(self.lines) > normal_area:
            self.lines = self.lines[-normal_area:]
        self.lines.extend(['\n'] * self.status_area)
        self.flush()

    def status_line(self, line):
        """Add line to the status area at the bottom of the status file.

        This area is set by set_status_area_height, and scrolls independently
        of the rest of the StatusFile.

        :param line: The line to add to the status area.
        """
        if not line.endswith('\n'):
            line += '\n'
        if not self.status_area:
            raise ValueError("no status area defined")
        self.lines.pop(-self.status_area)
        self.lines.append(line)
        self.flush()


class WriteToLineAdapter(object):
    """Adapt calls to write() to calls to line() on another object.

    when the content of write() is not \n terminated, \n is added.
    """

    def __init__(self, target):
        """Create a WriteToLineAdapter forwarding to target."""
        self.target = target

    def write(self, data):
        if data and data[-1] != '\n':
            data += '\n'
        self.target.line(data)
