# -*- mode: python; coding: utf-8 -*-
# vim:smartindent cinwords=if,elif,else,for,while,try,except,finally,def,class:ts=4:sts=4:sta:et:ai:shiftwidth=4
#
# Copyright © 2003,2004 Colin Walters <walters@verbum.org>
# Copyright © 2004 Canonical Ltd.
#   Author: Robert Collins <robertc@robertcollins.net>
# Copyright © 2003, 2005 Walter Landry
# Copyright © 2008 Canonical Ltd.
#   Author: Tim Penhey <tim@penhey.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging
import os

from optparse import OptionParser


DEFAULT_CONFIG_FILES = (
    '/etc/pqm.conf',
    '~/.pqm.conf',
    '~/.arch-pqm.conf',
    '~/.tla-pqm.conf',
    )


def normalise_filename(filename):
    """Expand the username and return an absolute path."""
    return os.path.abspath(os.path.expanduser(filename))


def default_pqm_config_files():
    """Return the default config files taking into account the environment."""
    if 'PQM_CONFIG' in os.environ:
        file_names = [os.environ['PQM_CONFIG']]
    else:
        file_names = [
            normalise_filename(filename) for filename in DEFAULT_CONFIG_FILES]
    return file_names


def increase_logging_level(option, opt, value, parser):
    """Make the logging more verbose."""
    current_level = parser.values.loglevel
    parser.values.loglevel = max(logging.DEBUG, current_level - 10)


def decrease_logging_level(option, opt, value, parser):
    """Make the less more verbose."""
    current_level = parser.values.loglevel
    parser.values.loglevel = min(logging.CRITICAL, current_level + 10)


def set_custom_config_file(option, opt, value, parser):
    """Override the default configfile_names."""
    if not parser.values.custom_config:
        parser.values.custom_config = True
        parser.values.configfile_names = []
    parser.values.configfile_names.append(normalise_filename(value))


def parse_command_line(args):
    parser = OptionParser()
    parser.set_defaults(
        configfile_names=default_pqm_config_files(),
        cron_mode=False,
        custom_config=False,
        debug_mode=False,
        keyring=None,
        loglevel=logging.WARN,
        no_act=False,
        no_log=False,
        read_mode=False,
        run_mode=False,
        print_report=False,
        show_verison=False,
        verify_sigs=True,
        queuedir=None
        )
    parser.add_option("-v", "--verbose", action="callback",
                      callback=increase_logging_level,
                      help="Display extra information")
    parser.add_option("-q", "--quiet", action="callback",
                      callback=decrease_logging_level,
                      help="Display less information")
    parser.add_option("-c", "--config", action="callback", metavar="FILE",
                      callback=set_custom_config_file, type="string",
                      help="Parse configuration info from FILE")
    parser.add_option("-d", "--debug", action="store_true", dest="debug_mode",
                      help="Output information to stdout as well as log")
    parser.add_option('--no-log', action="store_true",
                      help="Don't write information to log file")
    parser.add_option('-n', '--no-act', action="store_true",
                      help="Don't actually perform changes")
    parser.add_option('-r', '--read', action="store_true", dest="read_mode",
                      help="Read a request from stdin")
    parser.add_option("--run", action="store_true", dest="run_mode",
                      help="Process queue")
    parser.add_option("--cron", action="store_true", dest="cron_mode",
                      help=("Specify that run is being done from cron"
                            " (used with --run)"))
    parser.add_option("--report", action="store_true", dest="print_report",
                      help="Print patch report (used with --run)")
    parser.add_option("--no-verify", action="store_false", dest="verify_sigs",
                      help="Don't verify signatures")
    parser.add_option("--queuedir", metavar="DIR",
                      help="Perform first-time configuration")
    parser.add_option("--keyring", metavar="FILE",
                      help="Use the specified GPG keyring")
    parser.add_option("--version", action="store_true", dest="show_version",
                      help="Print the software version and exit")
    return parser.parse_args(args)
