# -*- mode: python; coding: utf-8 -*-
# vim:smartindent cinwords=if,elif,else,for,while,try,except,finally,def,class:ts=4:sts=4:sta:et:ai:shiftwidth=4
#
# Copyright © 2004, 2008 Canonical Ltd.
#	Author: Robert Collins <robertc@robertcollins.net>
#	Author: Tim Penhey <tim@canonical.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import smtplib

from pqm.errors import PQMCmdFailure, PQMException
from pqm.lockfile import LockFile
from pqm.queue.email import EmailQueue
from pqm.queue.list import ListQueue
from pqm.queue.lp import LaunchpadQueue


def get_script_queue(queuedir, logger, branch_spec_handler, configp,
                     options):
    """Determine the type of queue from the config."""
    return ListQueue([
        EmailQueue(queuedir, logger, branch_spec_handler, configp, options),
        # Talk to launchpad, use a common stop.patch as LP doesn't have a queue
        # offline mode yet.
        LaunchpadQueue(queuedir, configp, branch_spec_handler, logger),
        ])


def do_mkdir(name, logger, no_act=False):
    if os.access(name, os.X_OK):
        return
    try:
        logger.info('Creating directory "%s"' % (name))
    except:
        pass
    if not no_act:
        os.mkdir(name)


class PatchQueueManager(object):
    """The PatchQueueManager controls the core processing."""

    branch_configuration = None

    # There are four directories that the manager cares about:
    #  * queue_dir - the primary directory from which the other three
    #                default to being under.
    #  * work_dir - where the merging or commands takes place
    #  * log_dir - where the script execution log files are kept
    #  * control_dir - where the lockfile, and pqm log file is kept
    queue_dir = None
    work_dir = None
    log_dir = None
    control_dir = None

    def __init__(self, queue_dir, config_parser, options, logger):
        self.queue_dir = queue_dir
        self.config_parser = config_parser
        self.options = options
        self.logger = logger

        self.work_dir = self._dir_from_option('workdir', 'workdir')
        self.log_dir = self._dir_from_option('logdir', 'logs')
        self.control_dir = os.path.join(self.queue_dir, 'pqm')

        self.mail_reply = config_parser.get_boolean_option(
            'DEFAULT', 'mail_reply', True)
        self.mail_server = config_parser.get_option(
            'DEFAULT', 'mail_server', 'localhost')
        myname = config_parser.get_option(
            'DEFAULT', 'myname', 'Arch Patch Queue Manager')
        self.from_address = config_parser.get(
            'DEFAULT', 'from_address', None)
        if self.from_address is not None:
            self.nice_from_address = '%s <%s>' % (myname, self.from_address)

    def _dir_from_option(self, option, default):
        """Get the option from the config with a sensible default."""
        default_dir = os.path.join(self.queue_dir, default)
        configured_dir = self.config_parser.get_option(
            'DEFAULT', option, default_dir)
        return os.path.abspath(os.path.expanduser(configured_dir))

    def make_directories(self):
        """Make sure the directories exist."""
        do_mkdir(self.queue_dir, self.logger, self.options.no_act)
        do_mkdir(self.work_dir, self.logger, self.options.no_act)
        do_mkdir(self.log_dir, self.logger, self.options.no_act)
        do_mkdir(self.control_dir, self.logger, self.options.no_act)

    def run(self, script_queue):
        lockfile = LockFile(
            os.path.join(self.control_dir, 'pqm.lock'), self.logger,
            self.options.no_act, self.options.cron_mode)
        lockfile.acquire()
        try:
            self.do_run_mode(script_queue)
        finally:
            lockfile.release()

    def do_run_mode(self, script_queue):
        """Run through the script queue until empty."""
        (goodscripts, badscripts) = ([], [])

        script = script_queue.next_script()
        while script is not None:
            self.run_one_script(script, goodscripts, badscripts)
            script = script_queue.next_script()

        if self.options.print_report:
            for (patchname, logname) in goodscripts:
                print "Patch: " + patchname
                print "Status: success"
                print "Log: " + logname
                print
            for (patchname, logname) in badscripts:
                print "Patch: " + patchname
                print "Status: failure"
                print "Log: " + logname
                print

    def run_one_script(self, script, goodscripts, badscripts):
        # FIXME: Test it!
        logname = None
        try:
            success = False
            result = None
            try:
                script_id = script.script_id()
                self.logger.info('trying script ' + script_id)
                logname = os.path.join(self.log_dir, os.path.basename(
                        script_id) + '.log')

                script.pre_run_hook()

                output = []
                failedcmd=None

                result = script.run()
                successes = result.successes
                unrecognized = result.unrecognized
                output = result.output
                self.logger.info('successes: %s' % (successes,))
                self.logger.info('unrecognized: %s' % (unrecognized,))
                success = True
                goodscripts.append((script_id, logname))
            except PQMCmdFailure, e:
                badscripts.append((script_id, logname))
                successes = e.goodcmds
                failedcmd = e.badcmd
                output = e.output
                unrecognized=[]
            except PQMException, e:
                badscripts.append((script_id, logname))
                successes = []
                failedcmd = []
                output = [str(e)]
                unrecognized=[]
            except Exception, e:
                # catch all to ensure we get some output in uncaught failures
                output = [str(e)]
                raise
            if result is not None and result.report and self.mail_reply:
                self.send_mail_reply(success, successes, unrecognized,
                                     script.getSender(), failedcmd, output,
                                     script)
            else:
                self.logger.info('not sending mail success notification.')
        finally:
            log_list(logname, output)
            script.post_run_hook()

    def send_mail_reply(self, success, successes, unrecognized,
                        to_address, failedcmd, output, script):
        if success:
            retmesg = mail_format_successes(
                successes, "Command was successful.", unrecognized)
            if len(successes) > 0:
                statusmsg='success'
            else:
                statusmsg='no valid commands given'
        else:
            retmesg = mail_format_successes(
                successes, "Command passed checks, but was not committed.",
                unrecognized)
            retmesg += "\n%s" % failedcmd
            retmesg += '\nCommand failed!'
            retmesg += '\n' + script.format_output(output)
            statusmsg='failure'
        server = smtplib.SMTP(self.mail_server)
        server.sendmail(
            self.from_address,
            [to_address],
            'From: %s\r\nTo: %s\r\nSubject: %s\r\n\r\n%s\n' % (
                self.nice_from_address, to_address, statusmsg, retmesg))
        server.quit()


def mail_format_successes(successes, command_msg, unrecognized):
    retmesg = []
    for success in successes:
        retmesg.append('> ' + success)
        # Do we really want the command_msg for every success?
        retmesg.append(command_msg)
        # And I'm almost certain that we don't want these for each one.
        for line in unrecognized:
            retmesg.append('> ' + line)
            retmesg.append('Unrecognized command.')
    return '\n'.join(retmesg)


def log_list(logname, list):
    if logname is None:
        return
    f = open(logname, 'w')
    for l in list:
        f.write(l)
    f.close()

