# -*- mode: python; coding: utf-8 -*-
# vim:smartindent cinwords=if,elif,else,for,while,try,except,finally,def,class:ts=4:sts=4:sta:et:ai:shiftwidth=4
#
# Copyright © 2010 Canonical Ltd.
#	Author: Robert Collins <robert@canonical.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = [
    'get_launchpad',
    'LaunchpadQueue',
    ]

import copy
import os.path
import time

try:
    import launchpadlib.credentials
    import launchpadlib.launchpad
    try:
        from launchpadlib.uris import EDGE_SERVICE_ROOT
    except ImportError:
        EDGE_SERVICE_ROOT = launchpadlib.launchpad.EDGE_SERVICE_ROOT
    lp_service_root = os.environ.get('LAUNCHPAD_API', EDGE_SERVICE_ROOT)
except ImportError:
    launchpadlib = None

from pqm.queue import Queue
from pqm import script

def get_launchpad(configp):
    # NB: may want to preserve a single instance to avoid many connections
    # in complex scripts.
    if configp is None:
        keyfile = None
    else:
        keyfile = configp.get_option('DEFAULT', 'lp_api_keyfile', None)
    if not keyfile or launchpadlib is None:
        return None
    else:
        credentials = launchpadlib.credentials.Credentials()
        credentials.load(file(keyfile, "r"))
        from pqm import workdir
        cachedir = os.path.join(workdir, '.pqmcache')
        if not os.path.isdir(cachedir):
            os.makedirs(cachedir, mode=0700)
        return launchpadlib.launchpad.Launchpad(credentials, lp_service_root,
            cachedir)


class LaunchpadQueue(Queue):
    """A queue that reads from launchpad.

    Note that all branches in a LP project will have the same 'project' in the 
    PQM web UI - the configured projects get unioned as part of the lookup 
    process.
    
    :ivar controldir: A control dir which a stop.patch file can be written to
        to shutdown the queue.
    :ivar configp: A ConfigParser object to get configuration data from.
    :ivar rev_optionhandler: A BranchSpecOptionHandler describing the branches
        to read queues from.
    :ivar launchpad: A launchapd instance to talk to launchpad with.
    :ivar _items: A cached list of merges; when empty lp is queried on the
        first lookup.
    :ivar lp_projects: Launchpad project link -> project objects.
    :ivar projects: Launchpad project link -> configured projects in pqm.conf.
    :ivar logger: The logger to log with.
    """

    def __init__(self, controldir, configp, rev_optionhandler, logger):
        """Create a LaunchpadQueue controlled by controldir/stop.patch."""
        self.controldir = controldir
        self.configp = configp
        # We mutate the config; deep copy so that EmailQueue etc are not
        # impacted.
        self.rev_optionhandler = copy.deepcopy(rev_optionhandler)
        self.logger = logger
        self.launchpad = get_launchpad(configp)
        if self.launchpad is None:
            return
        # Look up amongst the configured branches all the branches on
        # launchpad, to get lp queues. Note that branch prefix specs are
        # not yet supported.
        branch_urls = []
        lookup_specs = {} # what spec did we look branches up at
        for key, spec in self.rev_optionhandler._specs.items():
            branch_urls.append(key)
            branch_urls.append(spec.get('published_at'))
            branch_urls.append(spec.get('publish_to'))
            lookup_specs[spec.get('published_at')] = key
            lookup_specs[spec.get('publish_to')] = key
            spec['commiters'] = None # Will be handled via LP permissions.
        branch_urls = list(set(branch_urls))
        # branch_urls = branch_urls[:10] lookup less for testing.
        branches = {}
        # Workound 414 error when asking for all urls at once
        count = len(branch_urls)
        per_query = 40
        segment_count = count / per_query + 1
        for index in range(segment_count):
            segment = branch_urls[index * per_query: (index + 1) * per_query]
            if not segment:
                continue
            # getByUrls is returning a dict of string->dict not string->Entry;
            # fail.
            # new_branches = self.launchpad.branches.getByUrls(urls=segment)
            # branches.update(new_branches)
            for url in segment:
                branches[url] = self.launchpad.branches.getByUrl(url=url)
        self.lp_branches = {}
        lp_projects = {}
        projects = {}
        for conf_url, branch in branches.items():
            if branch is not None:
                # Make the branches canonical location be the config lookup in
                # the branch specs, so that precommit lookups etc will work without
                # lots of magic later.
                if conf_url in lookup_specs:
                    self.rev_optionhandler._specs[lookup_specs[conf_url]]['published_at'] = branch.bzr_identity
                self.lp_branches[branch.self_link] = branch
                project = branch.project
                if project is not None:
                    lp_projects[project.self_link] = project
                    try:
                        target_config = self.rev_optionhandler.get_target_config(conf_url)[1]
                    except KeyError:
                        # Not a configured url per se
                        projects.setdefault(project.self_link, set())
                    else:
                        projects.setdefault(project.self_link, set()).add(target_config['project'])
                        projects[project.self_link].discard(None)
        self.lp_projects = lp_projects
        self.projects = projects
        self._items = None

    def is_processing_requests(self):
        """Is the queue currently processing requests?"""
        if self.launchpad is None:
            return False
        return not os.path.isfile("%s/stop.patch" % self.controldir)

    def next_script(self):
        if self.launchpad is None or not self._query_lp():
            return None
        return self._items.pop(0)

    def items(self):
        if self.launchpad is None:
            return []
        return list(self._query_lp())

    def _query_lp(self):
        if self._items is not None:
            # check for none so that when we successfully merge the last item
            # in the queue we don't query for more: until we affirm that things
            # pass, querying for more until the scanner notes the merge was
            # completed would be a Infinite Loop.
            return self._items
        items = []
        for project in self.lp_projects.values():
            mps = project.getMergeProposals(status=['Queued'])
            for mp in mps:
                if mp.target_branch.self_link not in self.lp_branches:
                    # Perhaps we should log this as being unhandled?
                    continue
                items.append((mp.queue_position,
                    LPScript(mp, self.projects[project.self_link],
                    self.rev_optionhandler, self.configp, self.logger)))
                script = 1
                # a merge to handle
        items.sort()
        self._items = [item[1] for item in items]
        return self._items


class LPScript(script.Script):
    """A script that was loaded from a launchpad merge proposal.
    
    :ivar merge_proposal: The merge proposal.
    :ivar projects: The projects to show under in the PQM web UI.
    :ivar _branch_spec_handler: Branch config looker-uper.
    :ivar _configp: The raw config parser.
    :ivar signing_email: None (set because the script Command objects look this
        up - but its used for access control tasks that are not used with lp,
        thus we set it to None).
    :ivar logger: Logger for logging with.
    :ivar _sender: The sender.
    """

    def __init__(self, merge_proposal, projects, branch_spec_handler, configp, logger):
        script.Script.__init__(self, logger)
        self.merge_proposal = merge_proposal
        self.projects = projects
        self._branch_spec_handler = branch_spec_handler
        self._configp = configp
        self.signing_email = None
        self.logger = logger
        self._sender = self.merge_proposal.queuer.display_name

    def getProjects(self): # UI uses this
        return self.projects

    def getSender(self): # UI uses this
        return self._sender

    def getSubject(self): # UI uses this
        # XXX: Extend to permit multiple line commmit messages.
        # XXX: Extend to permit a format rule or something, as LP probably does not
        # want the leading queuer.name that bzr uses.
        mp = self.merge_proposal
        return u"(%s) %s (%s)" % (mp.queuer.name, mp.commit_message,
                    mp.source_branch.owner.display_name)

    def getCommands(self): # UI uses this
        return [
            script.MergeCommand(self, self._branch_spec_handler, self._configp,
                self.merge_proposal.source_branch.bzr_identity,
                self.merge_proposal.target_branch.bzr_identity)
            ]

    def asctime(self): # UI uses this
        return "%s UTC" % time.asctime(
            self.merge_proposal.date_queued.utctimetuple())

    def isCommand(self, line):
        return True

    def script_id(self):
        """What is the unique id of this script?"""
        return "lp-" + self.merge_proposal.self_link.split('/')[-1]
