# -*- mode: python; coding: utf-8 -*-
# vim:smartindent cinwords=if,elif,else,for,while,try,except,finally,def,class:ts=4:sts=4:sta:et:ai:shiftwidth=4
#
# Copyright ©  2009 Robert Collins
#	Author: Robert Collins <robertc@robertcollins.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""A TestResult that generates a simple one line summary of activity."""

import math
import unittest

from subunit import PROGRESS_PUSH, PROGRESS_POP, PROGRESS_SET, PROGRESS_CUR
import subunit.progress_model

class TestStatusSummary(unittest.TestResult):
    """A test result that works with pqm's StatusFile.
    
    This shows errors as they happen and overall progress counters in the
    status area.
    """

    def __init__(self, status):
        """Create a TestStatusSummary outputting to status."""
        self.super = super(TestStatusSummary, self)
        self.super.__init__()
        self.status = status
        status.set_status_area_height(1)
        status.status_line('No subunit data.\n')
        self.state = 'before' # can be Running, or Finished, as well.
        self.current_test_id = None
        self.progress_model = subunit.progress_model.ProgressModel()
        self.total_tests = 0

    def _call_maybe(self, method_name, *params):
        """Call method_name on super if present.
        
        This is used to guard newer methods which older pythons do not
        support. While newer clients won't call these methods if they don't
        exist, they do exist on this object, and thus they may be called
        by any client.

        :param method_name: The name of the method to call.
        :param *params: Parameters to pass to method_name.
        :return: The result of super.method_name(*params), if it exists, and
            None otherwise.
        """
        method = getattr(self.super, method_name, None)
        if method is None:
            return
        return method(*params)
    
    def set_current_test(self, test):
        if test is None:
            self.current_test_id = None
        else:
            self.current_test_id = test.id()
        self.update()

    def update(self):
        """Update the status file with our current status."""
        if self.state == 'before':
            self.status.status_line('No subunit data.\n')
            return
        elements = [self.state]
        if self.progress_model.width() == 0:
            # No fraction known
            percent = 0
        else:
            percent = 100 * self.progress_model.pos() / float(
                self.progress_model.width())
            percent = math.ceil(percent)

        elements.append('[')
        elements.append('%d%%' % percent)
        if self.testsRun:
            elements.append('%d test(s)' % self.testsRun)
        if self.failures:
            elements.append('%d failure(s)' % len(self.failures))
        if self.errors:
            elements.append('%d errors(s)' % len(self.errors))
        elements.append(']')
        if self.state != 'Finished':
            elements.append('Current test: %s' % self.current_test_id)
        self.status.status_line(' '.join(elements) + '\n')

    def startTest(self, test):
        result = self.super.startTest(test)
        self.set_current_test(test)
        return result

    def startTestRun(self):
        result = self._call_maybe("startTestRun")
        self.state = 'Running'
        self.update()
        return result

    def stopTest(self, test):
        self.progress_model.advance()
        result = self.super.stopTest(test)
        self.set_current_test(None)
        return result

    def stopTestRun(self):
        result = self._call_maybe("stopTestRun")
        self.state = 'Finished'
        self.update()
        return result

    def addError(self, test, err):
        result = self.super.addError(test, err)
        self.status.line("Error %s\n" % test.id())
        for line in self.errors[-1][1]:
            self.status.line(line)
        self.status.line('\n')
        self.update()
        return result

    def addFailure(self, test, err):
        result = self.super.addFailure(test, err)
        self.status.line("Failure %s\n" % test.id())
        for line in self.failures[-1][1].splitlines():
            self.status.line(line + '\n')
        self.status.line('\n')
        self.update()
        return result

    def addSuccess(self, test):
        self.status.line("Success %s\n" % test.id())
        self.update()
        return self.super.addSuccess(test)

    def addSkip(self, test, reason):
        self.status.line("Skipped test %s: %s\n" % (test.id(), reason))
        self.update()
        return self._call_maybe("addSkip", test, reason)

    def addExpectedFailure(self, test, err):
        self.status.line("Expected failure %s\n" % test.id())
        self.update()
        return self._call_maybe("addExpectedFailure", test, err)

    def addUnexpectedSuccess(self, test):
        self.status.line("Unexpected success %s\n" % test.id())
        self.update()
        return self._call_maybe("addUnexpectedSuccess", test)

    def progress(self, offset, whence):
        result = self._call_maybe("progress", offset, whence)
        if whence == PROGRESS_PUSH:
            self.progress_model.push()
        elif whence == PROGRESS_POP:
            self.progress_model.pop()
        elif whence == PROGRESS_SET:
            self.progress_model.set_width(offset)
        else:
            self.progress_model.adjust_width(offset)
        return result

    def wasSuccessful(self):
        return self.super.wasSuccessful()

    def stop(self):
        return self.super.stop()

    def time(self, a_datetime):
        # PQM expects real-time tests always, so no special handling.
        return self._call_maybe("time", a_datetime)
