# -*- mode: python; coding: utf-8 -*-
# vim:smartindent cinwords=if,elif,else,for,while,try,except,finally,def,class:ts=4:sts=4:sta:et:ai:shiftwidth=4
#
# Copyright ©  2004 Canonical Ltd.
#	Author: Robert Collins <robertc@robertcollins.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import cgi
from datetime import datetime
import email
import errno
import re
import os
import select
import shutil
import StringIO
import subprocess
import sys
import time
import urllib

import config_manager
import subunit.test_results

from pqm.errors import PQMCmdFailure, PQMException, PQMTlaFailure


def read_email(logger, file=None):
    """Read an email and check it has the required structure for commands."""
    if not file:
        file = sys.stdin
    msg = email.message_from_file(file)
    sender = msg['From']
    logger.info("recieved email from %s", sender)
    subject = msg['Subject']
    if not sender:
        raise PQMException(None, "No From specified")
    if (not subject) or subject=='':
        raise PQMException(sender, "No Subject specified")
    text = None
    sig = None
    if msg.is_multipart():
        print ("Multipart support is known buggy. Patches, or better still"
               "unittest tests with or without patches solicited.")
        parts = msg.get_payload()
        if not len(parts) == 2:
            raise PQMException(sender,
                        "Multipart message must have exactly two parts")
        if not parts[0].get_content_type() == 'text/plain':
            raise PQMException(sender,
                        "First part of multipart message must be text/plain")
        if not parts[1].get_content_type() == 'application/pgp-signature':
            raise PQMException(sender,
                        "Second part of multipart message must be"
                        " application/pgp-signature")
        return (sender,
                subject,
                parts[0].get_payload(),
                parts[1].get_payload())
    else:
        return (sender, subject, msg.get_payload(), None)


class Script(object):
    """A script for the Patch Queue Manager to process.

    A script will contain one or more `Command`s.
    """

    def __init__(self, logger):
        # Default to debug output
        self.debug = True
        self.logger = logger

    def pre_run_hook(self):
        """This hook is called before the script is run."""

    def post_run_hook(self):
        """This hook is called after the script is run.

        This method is called regardless of whether or not run raised
        an exception.
        """

    def format_output(self, output):
        """Format some output lines."""
        # XXX: This should do a subunit filter on output to strip
        # success messages.
        if not self.debug:
            rawresult = ('Last 20 lines of log output:' +
                ''.join(output[-20:]))
        else:
            rawresult = ('All lines of log output:' + 
                ''.join(output))
        stream = StringIO.StringIO(rawresult)
        del rawresult
        output_file = StringIO.StringIO()
        client = subunit.TestProtocolClient(output_file)
        # Launchpad may want to keep zope testing tagged tests.
        client = subunit.test_results.TestResultFilter(client)
        parser = subunit.ProtocolTestCase(stream, passthrough=output_file)
        parser.run(client)
        del stream
        output_file.seek(0)
        result = StringIO.StringIO()
        # discard repeated time: statements. Iz subunit bug.
        last_was_time = False
        for line in output_file.readlines():
            is_time = line.startswith('time: ')
            if not (last_was_time and is_time):
                result.write(line)
            last_was_time = is_time
        return result.getvalue()

    def run(self):
        """Run this script.

        :return: (successes, unrecognized, output)
        """
        # reporting: reporting is framed as a reporting problem per-branch, and
        # scripts may address many branches; so things that fail early won't
        # report to a web service. If all the commands are set to hide their
        # output, don't output anything at all.
        result = CommandResult([], [], [], False)
        for command in self.getCommands():
            command_result = command.run()
            result.report = result.report or command_result.report
            if command_result.report:
                result.successes.extend(command_result.successes)
                result.unrecognized.extend(command_result.unrecognized)
                result.output.extend(command_result.output)
        return result

    def script_id(self):
        """What is the unique id of this script?"""
        raise NotImplementedError(self.script_id)


class EmailScript(Script):
    """A command script generated from an email."""

    whitespace_re = re.compile('^\s*$')
    pgp_re = re.compile('^-----BEGIN PGP.*MESSAGE')
    pgp_end_re = re.compile('^-----BEGIN PGP SIG')
    # parser for merge recognition
    star_re = re.compile('^star-merge (\S+/\S+)\s+(\S+/\S+)\s*$')
    # parse matcher for the debug command
    debug_re = re.compile('^debug')

    def __init__(self, filename, logger, verify_sigs, submission_time,
                 branch_spec_handler, configp, keyring=None):
        """Create a script for a given file."""
        Script.__init__(self, logger)
        self.filename = filename
        self.verify_sigs = verify_sigs
        self.readComplete = False
        self.submission_time = submission_time
        self._branch_spec_handler = branch_spec_handler
        self._configp = configp
        self._keyring = keyring

    def pre_run_hook(self):
        """This hook is called before the script is run."""
        # Make sure that the message has been read, and that the
        # signature has been verified if needed.
        self._read()

    def post_run_hook(self):
        # ensure we always unlink the script file.
        os.unlink(self.filename)

    def _read(self):
        """Read in the script details."""
        #  This is cruft from the binary script. It actually does need to read
        #  the file twice with the current layering - yes thats funky, but not
        #  something I plan on breaking in an ad hoc manner right now. The
        #  read_email and email.message_from_file and verify_sig functions
        #  need rearranging and consolidating to allow this.
        details = read_email(self.logger, open(self.filename))
        (self.sender, self.subject, self.msg, self.sig) = details
        self.signing_email = None
        if self.verify_sigs:
            # TODO: Tim Penhey, 2008-05-28
            # Get rid of the imports from pqm module itself.
            from pqm import verify_sig
            sigid, siguid = verify_sig(self.sender,
                                       self.msg,
                                       self.sig,
                                       0,
                                       self.logger,
                                       self._keyring)
            self.signing_email = siguid
        self.msg = email.message_from_file(open(self.filename))
        self.sender = self.msg['From']
        self.logger.info ('sender: %s' , self.sender)
        self.logger.info ('subject: %s' , self.subject)
        self.readComplete = True

    def getProjects(self):
        """Return the set of projects this script will affect."""
        result = set()
        for command in self.getCommands():
            result.update(command.getProjects())
        return result

    def getSender(self):
        """Get the sender of the script."""
        if not self.readComplete:
            self._read()
        return self.sender

    def getSubject(self):
        """Get the subject of the script."""
        if not self.readComplete:
            self._read()
        return self.subject

    def getContent(self):
        """Get the raw content for folk that want it."""
        if not self.readComplete:
            self._read()
        return self.msg.get_payload()

    def getLines(self):
        """Get the lines in the script."""
        if not self.readComplete:
            self._read()
        to_skip = 0
        result = []
        for line in self.msg.get_payload().split('\n'):
            if to_skip:
                to_skip -= 1
                continue
            if self.whitespace_re.match(line):
                continue
            if self.pgp_re.match(line):
                to_skip = 2
                continue
            if self.pgp_end_re.match(line):
                break
            result.append(line)
        return result

    def getCommands(self):
        """Get the actual command lines from the script."""
        self.logger.info("parsing commands")
        result = []
        legacy_lines = []
        for line in self.getLines():
            if not self.isCommand(line):
                continue
            # identify and construct commands
            star_match = EmailScript.star_re.match(line)
            debug_match = EmailScript.debug_re.match(line)
            any_match = star_match or debug_match
            if any_match and legacy_lines:
                result.append(CommandRunner(self,
                                            self._branch_spec_handler,
                                            self._configp,
                                            legacy_lines,
                                            ))
                legacy_lines = []
            if star_match:
                result.append(MergeCommand(self,
                                           self._branch_spec_handler,
                                           self._configp,
                                           star_match.group(1),
                                           star_match.group(2)))
            elif debug_match:
                result.append(DebugCommand(self,
                                           self._branch_spec_handler,
                                           self._configp))
            else:
                legacy_lines.append(line)
        if legacy_lines:
            result.append(CommandRunner(self,
                                        self._branch_spec_handler,
                                        self._configp,
                                        legacy_lines))
        return result

    def asctime(self):
        """Return a string describing the time of the script."""
        return "%s UTC" % time.asctime(time.gmtime(self.getSubmissionTime()))

    def getSubmissionTime(self):
        """Return the time the script was submitted."""
        return self.submission_time

    def isCommand(self, line):
        """Return true if line looks like a valid command line."""
        # XXX: This is not ready yet, its still in CommandRunners guts.
        return True

    def script_id(self):
        """What is the unique id of this script?"""
        return "email-" + self.filename


class ConfigManagerTree(object):
    """A tree of branches managed via config-manager."""

    def __init__(self, path, config_url, override_mapper):
        """Construct the tree object."""
        self.path = path
        self.config_url = config_url
        self.override_mapper = override_mapper
        self.config = None

    def _read_config(self):
        """
        Reads and returns the current config-manager configuration for this
        tree.
        """
        try:
            stream = urllib.urlopen(self.config_url)
        except IOError,e:
            if e.errno == errno.ENOENT:
                raise PQMException(sender, "No such config '%s'" % self.config_url)
            else:
                raise PQMException(sender, "IOError opening configuration '%s'" % self.config_url)
        return config_manager.Config(fromStream=stream,
                                     override_mapper=self.override_mapper)

    def build(self):
        """Read the configuration and build the tree in the filesystem."""
        config = self._read_config()
        config.build(self.path)
        self.config = config

    def update(self):
        """
        Re-read the configuration and rebuild the tree in the filesystem if it
        has changed since the tree was built or last updated.
        """
        old_config = self.config
        config = self._read_config()
        if old_config is None or \
           config.get_entries() != old_config.get_entries():
            config.update(self.path)
        self.config = config


class Command(object):
    """A single command in a script.

    These are generated by the parser for specific messages - such as Script
    for the original format pqm messages.
    
    :ivar reported: Has the merge been reported (e.g. to a web service, bug
        tracker etc).
    :ivar launchpad: An instance of the launchpad web API. This is the initial
        web service based reporting mechanism - if you're looking at this code
        and want to report to another web based mechanism, please just refactor
        appropriately - it should be straight forward. If None, launchpad 
        isn't configured in the PQM config.
    """

    def __init__(self, script, branch_spec_handler, configp):
        """Construct a command.

        :param script: The Script object with user details etc for this command.
        :param branch_spec_handler: The branch specification handler for this
                                    PQM.
        :param configp: The ConfigParser for this PQM.
        """
        self.script = script
        self.branch_spec_handler = branch_spec_handler
        self.configp = configp
        self._url_override_mapper = None
        self.reported = False
        from pqm.queue import lp # circular dep.
        self.launchpad = lp.get_launchpad(self.configp)

    def asHTML(self):
        """Return an HTML representation of the command."""
        raise NotImplementedError(self.asHTML)

    def __eq__(self, other):
        # command top level objects have no intrinsic state.
        return isinstance(other, Command)

    def getBranchConfig(self, branch_spec):
        return self.branch_spec_handler.get_target_config(branch_spec)

    def getProjects(self):
        """Return the projects this Command will affect."""
        raise NotImplementedError(self.getProjects)

    def log_with_status(self, logger, message, *args):
        """Log message to the logfile, and to the status file."""
        logger.info(message, *args)
        self.branch_spec_handler.status.line(message % args)

    def _branch_name(self, branchspec):
        elements = branchspec.split('/')
        return elements[-1]

    def branch_from_config(self, config, sender, branch_list, fullpath):
        """build the config config in dir fullpath. Then
        find one of branch_list in the config, and
        return a (path, cm_tree) pair."""
        # TODO: probe the config path for the url to get -
        # config package isn't guaranteed arch-style.
        config_segments = config.split('/')
        config_path = config_segments.pop()
        done = False
        while not done:
            try:
                config_branch = '/'.join(config_segments)
                self.get_branch_handler(config_branch).make_local_dir(sender, config_branch, fullpath)
                done = True
            except PQMTlaFailure, e:
                if config_path == config or len(config_segments) == 1:
                    raise
                config_path = '/'.join((config_segments.pop(), config_path))
                if os.path.exists(fullpath):
                    shutil.rmtree(fullpath)
        cm_config_url = os.path.join(fullpath, config_path)
        cm_tree = ConfigManagerTree(path=fullpath, config_url=cm_config_url,
                                    override_mapper=self._get_url_override_mapper())
        cm_tree.build()
        if config_branch in branch_list:
            return (fullpath, cm_tree)
        for path, entry in cm_tree.config.get_entries().items():
            # probably want a url comparison that understands url parameters
            # TODO: but that can wait.
            if (entry.url in branch_list or
                (entry.url.startswith('arch://') and
                 entry.url[7:] in branch_list)):
                return (os.path.join(fullpath, path), cm_tree)
        raise PQMException(sender,
                           "Branch %s not found in config" % branch_list)

    def check_commit_regex(self, branch, config):
        """Confirm that commit message matches any regexp supplied in the
        configuration file.
        """
        if not config["commit_re"]:
            # No regexp, therefore accept anything
            return
        regex = config["commit_re"]
        if re.match(regex, self.commitmsg):
            # Regexp matched, accept the commitmsg
            return
        self.fail_operation("Commit message [%s] does not match commit_re [%s]"
                           % (self.commitmsg, regex))

    def fail_merge_proposal(self):
        """Mark a merge proposal as failed.
        
        Exists so that when LP bug https://bugs.launchpad.net/launchpad-code/+bug/561160
        is fixed, only one call site needs updatting.

        For now, work around that bug by setting the merge back to approved:
        once the issue is fixed it can be resubmitted.
        """
        if self.lp_mp is not None:
            self.lp_mp.setStatus(status='Approved')

    def fail_with_pqm_failure(self, line, e):
        """Fail an operation with a PQMTlaFailure.

        This will report to a web service if one is available and raises
        PQMCmdFailure afterwards.

        :param line: A command line that has failed (e.g. a merge command).
        :param e: a PQMTlaFailure or PQMCmdFailure to fail with.
        :raises PQMCmdFailure:
        """
        try:
            if self.lp_mp is not None:
                import core
                success_msg = core.mail_format_successes(self.successful, line,
                    [])
                output = self.script.format_output(self.output + [e.msg])
                message = "Successful steps %s\nFailure output:\n%s" % (success_msg, output)
                # XXX: subject should be 'the right thing for the thread' - how to
                # figure that out?
                self.lp_mp.createComment(content=message, subject="PQM failure")
                self.fail_merge_proposal()
                self.reported = True
        except KeyboardInterrupt:
            raise
        except Exception:
            # failures in talking to the web service should not break pqm:
            # swallow them.
            logger = self.script.logger
            logger.error("Webservice reporting failure", exc_info=1)
        if isinstance(e, PQMCmdFailure):
            raise e
        elif isinstance(e, PQMTlaFailure):
            raise PQMCmdFailure(self.script.getSender(), self.successful, line, self.output + e.output)
        else:
            raise PQMCmdFailure(self.script.getSender(), self.successful, line, self.output + [e.msg])

    def fail_operation(self, message):
        """Fail the operation with message message.

        This will report to a web service if one is available and raise
        PQMException.
        """
        try:
            if self.lp_mp is not None:
                # XXX: subject should be 'the right thing for the thread' - how to
                # figure that out?
                self.lp_mp.createComment(content=message, subject="PQM failure")
                self.fail_merge_proposal()
                self.reported = True
        except KeyboardInterrupt:
            raise
        except Exception:
            # failures in talking to the web service should not break pqm:
            # swallow them.
            logger = self.script.logger
            logger.error("Webservice reporting failure", exc_info=1)
        raise PQMException(self.script.getSender(), message)

    def check_target(self, branch, line):
        """Check that the sender is allowed to commit to torepo/to_revision"""
        # FIXME check gpg etc etc.
        try:
            self.getBranchConfig(branch)
        except KeyError:
            raise PQMCmdFailure(self.script.getSender(),
                                self.successful,
                                line,
                                ["Sender not authorised to commit to branch %s"
                                    % branch])

    def _success_output(self, command_name, start_time):
        """The success output strings with current and elapsed time."""
        elapsed_time = datetime.now() - start_time
        return ['\n', '%s succeeded at %s (%s)' % (
               command_name, time.strftime('%c'), elapsed_time), '\n']

    def cleanup_wd(self):
        """Cleans up all possible working dirs."""
        # TODO: Tim Penhey, 2008-05-28
        # Get rid of the imports from pqm module itself.
        from pqm import workdir
        self.log_with_status(self.script.logger, "cleaning working directory")
        for top in os.listdir(workdir):
            if top != '.pqmcache':
                self.rm_rf(os.path.join(workdir, top))
        for branch, value in self.branch_spec_handler._specs.items():
            possible_dir = value['build_dir']
            if not possible_dir:
                continue
            if not os.path.exists(possible_dir):
                continue
            for top in os.listdir(possible_dir):
                self.rm_rf(os.path.join(possible_dir, top))

    def do_merge(self, from_repo_revision, to_repo_revision, merge_name,
                 merge_method, line):
        # Record the start time for elapsed time calculation.
        start_time = datetime.now()
        sender = self.script.getSender()
        # Star-merge
        self.check_target(to_repo_revision, line)
        to_repo_revision, config = self.getBranchConfig(to_repo_revision)
        self.set_current_vcs(from_repo_revision, to_repo_revision)
        self.validate_revision(from_repo_revision)
        self.validate_revision(to_repo_revision)
        self.check_commit_regex(to_repo_revision, config)
        logger = self.script.logger
        logger.info("current cwd is %s", os.getcwd())
        logger.info("getting working dir for %s", to_repo_revision)
        dir, cm_tree = self.get_wd_and_cm_tree(sender, to_repo_revision, config)
        origdir = os.getcwd()
        self.log_with_status(logger, 'Executing %s %s at %s', merge_name,
            from_repo_revision, time.strftime('%c'))
        self.output += [
            '\n',
            'Executing %s %s at %s' % (merge_name, from_repo_revision,
                time.strftime('%c')),
            '\n',
            ]
        self.wrap_command(getattr(self.get_vcs(), merge_method), line, sender,
                                  from_repo_revision, dir)
        try:
            if cm_tree:
                # merge may have changed the config-manager config; update the tree
                cm_tree.update()
        except KeyboardInterrupt:
            raise
        except Exception, e:
            self.fail_operation(str(e))
        try:
            self.run_precommit(to_repo_revision, config, line, dir)
        except PQMCmdFailure, e:
            self.fail_with_pqm_failure(line, e)
        except KeyboardInterrupt:
            raise
        except Exception, e:
            self.fail_operation(str(e))
        os.chdir(origdir)
        self.log_with_status(logger, "success: %s", line)
        self.successful.append(line)
        self.output += self._success_output(merge_name, start_time)
        self.wrap_command(self.get_vcs().commit, "Commit merged code", sender,
            dir, self.commitmsg, to_repo_revision, config)
        # Now that we've committed to the branch, let's include the 
        # commit details in the output
        try:
            revno, commit_msg = self.get_vcs().latest_revision_info(sender, to_repo_revision)
            self.successful[-1] = ("%s\n" 
                            "Revision: %s \n"
                            "Commit Message: %s" 
                            % (line, revno, commit_msg))
            self.output += ["\n", "Revision Details:", "\n", 
                "Revision: %s" % revno, "\n", 
                "Commit Message: %s" % commit_msg, "\n"]
        except PQMTlaFailure:
            # If the latest_revision_info function isn't implemented
            pass
        if self.lp_mp is not None:
            # Launchpad notices merges 
            self.reported = True

    def get_arch_impl(self):
        # TODO: Tim Penhey, 2008-05-28
        # Get rid of the imports from pqm module itself.
        from pqm import ArXHandler, Baz1_1Handler, popen_noshell, TlaHandler
        if self.arch_impl is None:
            arch_impl = None
            (status, msg, output) = popen_noshell(self.arch_path, '--version')
            for line in output:
                if line.find('baz ') >= 0:
                    arch_impl = Baz1_1Handler()
                    break
                elif line.find('tla ') >= 0:
                    arch_impl = TlaHandler()
                    break
                elif line.find('ArX ') >= 0:
                    arch_impl = ArXHandler()
                    break
            if not arch_impl:
                # Log details - info because a bad url for all implementations
                # will take this code path but tla may not be installed: its
                # not an error to not have a vcs you aren't using missing from
                # the system.
                self.script.logger.info(
                    "Couldn't determine arch implementation.\n"
                    "please set arch_impl\n"
                    "(output %s)\n"
                    "(msg %s)\n" % (output, msg))
                # Exception for flow control
                raise PQMException(self.script.getSender(),
                    "Could not determine arch command.")
            self.arch_impl = arch_impl
        return self.arch_impl

    def get_branch_handler(self, branchspec):
        # TODO: Tim Penhey, 2008-05-28
        # Get rid of the imports from pqm module itself.
        # -- probe Bzr
        try:
            from pqm import Bazaar2Handler
            handler = Bazaar2Handler()
            # -- probe arx/tla/baz etc
            if handler.branch_exists(self.script.getSender(), branchspec):
                return handler
            handler = self.get_arch_impl()
        except PQMException:
            raise PQMException(self.script.getSender(),
                "Could not determine branch type for %r" % branchspec)
        return handler

    def get_wd(self, sender, branch, config):
        return self.get_wd_and_cm_tree(sender, branch, config)[0]

    def get_wd_and_cm_tree(self, sender, branch, config):
        # TODO: Tim Penhey, 2008-05-28
        # Get rid of the imports from pqm module itself.
        from pqm import groups, workdir
        dirpath = self._make_wd_path(workdir, branch)
        commiters = config['commiters']
        if commiters and not self.user_email in groups[commiters]:
            self.script.logger.error("%s is not permitted to commit to %s",
                                     self.user_email,
                                     branch)
            raise PQMException(sender,
                               "%s is not permitted to commit to %s" % (
                                    self.user_email, branch))
        possible_dir = config['build_dir']
        if possible_dir:
            dirpath = self._make_wd_path(possible_dir, branch)
        build_config = config['build_config']
        fullpath = self.prep_wd(sender, dirpath, branch, build_config)
        if build_config:
            branch_list = [branch]
            if config.get('published_at', None):
                branch_list.append(config['published_at'])
            return self.branch_from_config(build_config, sender, branch_list, fullpath)
        else:
            self.get_vcs().make_local_dir(sender, branch, fullpath)
            return (fullpath, None)

    def _make_wd_path(self, directory, branch):
        """Make the path to be used for branch in directory"""
        elements = branch.split('/')
        index = 0
        if elements[0] == '':
            index = 1
        result = os.path.join(directory, elements[index])
        if result[-1] == ':':
            result = result[:-1]
        return result

    def prep_wd(self, sender, dirpath, branch, config):
        branch_name = self._branch_name(branch)
        if not os.access(dirpath, os.W_OK):
            os.mkdir(dirpath)

        fullpath=os.path.join(dirpath, branch_name)
        if os.access(fullpath, os.W_OK):
            self.script.logger.error("Working dir already exists: " + fullpath)
            raise PQMException(sender, "Working dir already exists: " + fullpath)
        return fullpath

    def _get_url_override_mapper(self):
        """Get a URL override mapper from the config."""
        if self._url_override_mapper is not None:
            return self._url_override_mapper
        mapper = config_manager.URLMapper()
        if not self.configp.has_section('location overrides'):
            return mapper
        for target, source in self.configp.section_items('location overrides'):
            equal_pos = source.find('=')
            if (equal_pos > -1 and target.find('//') == -1 and
                source.startswith('//')):
                # : triggers separation, THANKS config-parser.
                target = target + ":" + source[:equal_pos]
                source = source[equal_pos + 1:]
                print target, source
            mapper.add_map(source, target)
        self._url_override_mapper = mapper
        return mapper

    def get_vcs(self):
        return self._vcs

    def validate_revision(self, branch):
        if not self.get_vcs().branch_exists(self.script.getSender(), branch):
            raise PQMCmdFailure(self.script.getSender(),
                                self.successful,
                                line,
                                self.output + self.get_vcs().last_error.output)

    def rm_rf(self, top):
        os.chmod(top, 0755)
        for root, dirs, files in os.walk(top, topdown=True):
            for name in files:
                if not os.path.islink(os.path.join(root, name)):
                    os.chmod(os.path.join(root, name), 0644)
            for name in dirs:
                if not os.path.islink(os.path.join(root, name)):
                    os.chmod(os.path.join(root, name), 0755)

        for root, dirs, files in os.walk(top, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                if os.path.islink(os.path.join(root, name)):
                    os.remove(os.path.join(root, name))
                else:
                    os.rmdir(os.path.join(root, name))

    def run(self):
        """Run this script.

        :return: A CommandResult
        """
        # current prelude code useful to all commands
        # may need to shrink or grow in the future.
        self.successful = []
        self.unrecognized = []
        self.output = []
        self.commitmsg = self.script.getSubject()
        self.user_email = self.script.signing_email
        self.sender = self.script.getSender()

    def run_in_dir(self, local_dir, method, *args):
        """call method after chdiring to local_dir, then chdir back."""
        orig_dir = os.getcwdu()
        try:
            os.chdir(local_dir)
            return method(*args)
        finally:
            os.chdir(orig_dir)

    def run_precommit(self, branch, config, line, dir):
        command = PrecommitCommand(
            self.script,
            self.branch_spec_handler,
            self.configp,
            config,
            line,
            dir)
        result = command.run()
        self.successful.extend(result.successes)
        self.unrecognized.extend(result.unrecognized)
        self.output.extend(result.output)
        # If the precommit succeeded, we're not fully reported even if it
        # reported a partial step.

    def set_current_vcs(self, branch, other_branch=None):
        branch_vcs = self.get_branch_handler(branch)
        if other_branch is not None:
            other_vcs = self.get_branch_handler(other_branch)
            if branch_vcs.__class__ != other_vcs.__class__:
                raise PQMException(self.script.getSender(),
                                   ["PQM Cannot merge between different VCS"
                                    "systems. '%s'(%s) and '%s'(%s) are different."
                                    % (branch, branch_vcs.__class__ ,
                                       other_branch, other_vcs.__class__)] + self.output)
        self._vcs = branch_vcs

    def wrap_command(self, command, line, sender, *args):
        try:
            self.output += command(sender, *args)
        except PQMTlaFailure, e:
            self.fail_with_pqm_failure(line, e)


class DebugCommand(Command):
    """Turns on debug mode."""

    def asHTML(self):
        return "debug"

    def getProjects(self):
        """Debug applies to no projects."""
        return set()

    def run(self):
        """Turn debug mode on."""
        super(DebugCommand, self).run()
        self.script.debug = True
        return CommandResult(['debug'], [], ['Debug mode enabled.'])


class MergeCommand(Command):
    """A merge request.
    """

    def __init__(self, script, branch_spec_handler, configp, from_branch, to_branch):
        """Construct a merge command.

        :param from_branch: The branch specification to merge from.
        :param to_branch: The branch specification to merge into.
        """
        super(MergeCommand, self).__init__(script, branch_spec_handler, configp)
        self.from_branch = from_branch
        self.to_branch = to_branch

    def asHTML(self):
        return cgi.escape("Merge %s %s" % (self.from_branch, self.to_branch))

    def __eq__(self, other):
        return (super(MergeCommand, self).__eq__(other) and
                isinstance(other, MergeCommand) and
                self.from_branch == other.from_branch and
                self.to_branch == other.to_branch)

    def getProjects(self):
        b, config = self.getBranchConfig(self.to_branch)
        project = config['project']
        if project is not None:
            return set([project])
        else:
            return set()

    def run(self):
        super(MergeCommand, self).run()
        self.cleanup_wd()
        self.lp_mp = None
        if self.launchpad is None:
            self.from_branch_lp = None
            self.to_branch_lp = None
        else:
            # LPLib doesn't return branch objects from getByUrls
            #lplib_branches = self.launchpad.branches.getByUrls(
            #    urls=[self.from_branch, self.to_branch])
            ## Branch objects in LP
            #self.from_branch_lp = lplib_branches[self.from_branch]
            #self.to_branch_lp = lplib_branches[self.to_branch]
            # Branch objects in LP
            self.from_branch_lp = self.launchpad.branches.getByUrl(
                url=self.from_branch)
            self.to_branch_lp = self.launchpad.branches.getByUrl(
                url=self.to_branch)
            # Merge proposal in LP
            for merge_proposal in self.from_branch_lp.landing_targets:
                if merge_proposal.target_branch.self_link == self.to_branch_lp.self_link:
                    self.lp_mp = merge_proposal
                    break
        line = 'merge %s %s' % (self.from_branch, self.to_branch)
        try:
            self.do_merge(from_repo_revision=self.from_branch,
                          to_repo_revision=self.to_branch,
                          merge_name='star-merge',
                          merge_method="do_star_merge",
                          line=line)
        except KeyboardInterrupt:
            raise
        except PQMException, e:
            self.fail_with_pqm_failure(line, e)
        except PQMCmdFailure:
            # Ditto
            raise
        except Exception, e:
            # Something unexpected has gone wrong, make sure we report it.
            self.fail_operation("Exception processing merge: %s" % e)
        return CommandResult(self.successful, self.unrecognized, self.output, not self.reported)


class PrecommitCommand(Command):
    """A pre-commit command."""

    def __init__(self, script, branch_spec_handler, configp, config, line, dir):
        """Construct a merge command.

        :param config: The branch config to use for the precommit logic.
        :param line: The parsed line(s) that triggered the precommit hook.
        :param dir: The directory to run the precommit check in.
        """
        super(PrecommitCommand, self).__init__(script, branch_spec_handler, configp)
        self.config = config
        self.line = line
        self.dir = dir

    def asHTML(self):
        return cgi.escape("Pre-commit tests")

    def __eq__(self, other):
        return (super(PrecommitCommand, self).__eq__(other) and
                isinstance(other, PrecommitCommand) and
                self.from_branch == other.from_branch and
                self.to_branch == other.to_branch)

    def getProjects(self):
        b, config = self.getBranchConfig(self.to_branch)
        project = config['project']
        if project is not None:
            return set([project])
        else:
            return set()

    def iter_child_lines(self, pipe):
        '''process the output of a child into lines and yield them.'''
        read_set = []
        read_set.append(pipe.stdout)
        read_set.append(pipe.stderr)

        out_data = err_data = ''
        while read_set:
            rlist, wlist, xlist = select.select(read_set, [], [])

            if pipe.stdout in rlist:
                out_chunk = os.read(pipe.stdout.fileno(), 1024)
                if out_chunk == '':
                    pipe.stdout.close()
                    read_set.remove(pipe.stdout)
                out_data += out_chunk
                while '\n' in out_data:
                    pos = out_data.find('\n')
                    yield "out", out_data[:pos+1]
                    out_data = out_data[pos+1:]

            if pipe.stderr in rlist:
                err_chunk = os.read(pipe.stderr.fileno(), 1024)
                if err_chunk == '':
                    pipe.stderr.close()
                    read_set.remove(pipe.stderr)
                err_data += err_chunk
                while '\n' in err_data:
                    pos = err_data.find('\n')
                    yield "err", err_data[:pos+1]
                    err_data = err_data[pos+1:]

            select.select([],[],[],.1) # give a little time for buffers to fill

    def run(self):
        super(PrecommitCommand, self).run()
        # Circular... TODO: fix.
        from pqm import resultsummary, WriteToLineAdapter
        hook = self.config['precommit_hook']
        if not hook:
            # TODO: Tim Penhey, 2008-05-28
            # Get rid of the imports from pqm module itself.
            from pqm import precommit_hook
            hook = precommit_hook
        if hook:
            start_time = datetime.now()
            self.script.logger.info("running precommit hook: %s" % (hook,))
            self.output.extend(['\n', 'Executing pre-commit hook %s at %s' % (hook, time.strftime('%c')), '\n'])
            # execute the child keeping stdout/stderr separate so we don't get
            # partial-lines mixed with each other.
            child = subprocess.Popen(hook, stderr=subprocess.PIPE,
                stdout=subprocess.PIPE, stdin=subprocess.PIPE, close_fds=True,
                cwd=self.dir, shell=True)
            child.stdin.close()
            status = self.branch_spec_handler.status
            results = resultsummary.TestStatusSummary(status)
            test = subunit.TestProtocolServer(results, WriteToLineAdapter(status))
            results.startTestRun()
            for source, line in self.iter_child_lines(child):
                if source == 'out':
                    if line and line[-1] != '\n':
                        line = line + '\n'
                    test.lineReceived(line)
                else:
                    status.line(line)
                self.output.append(line)
            results.stopTestRun()
            ecode = child.wait()
            if not ((ecode is None) or (ecode == 0)):
                self.output.append('\npre-commit hook failed with error code %d at %s\n' % (ecode - 255, time.strftime('%c')))
                raise PQMCmdFailure(self.sender, self.successful, self.line, self.output)
            self.output += self._success_output('pre-commit hook', start_time)
        return CommandResult(self.successful, self.unrecognized, self.output, not self.reported)


class CommandResult(object):
    """Result of running a command.
    
    :ivar successes: A list of the successful commands.
    :ivar unrecognized: A lsit of the unrecognized commansd.
    :ivar output: Output to show the user.
    :ivar report: If True, report on this command, if False do not report.
    """

    def __init__(self, successes, unrecognized, output, report=True):
        """Create a CommandResult. See the class docstring for details."""
        self.successes = successes
        self.unrecognized = unrecognized
        self.output = output
        self.report = report


class CommandRunner(Command):
    """This runs commands as lines one at a time."""
    replay_re = re.compile('^replay (\S+/\S+)\s+(\S+/\S+)\s*$')
    repo_cache_re = re.compile('^repo-cache-revision (\S+/\S+)\s*$')
    repo_uncache_re = re.compile('^repo-uncache-revision (\S+/\S+)\s*$')
    tag_re = re.compile('^tag (\S+/\S+)\s+(\S+/\S+)\s*$')
    make_repo_re = re.compile('^make-repo (\S+)\s+(\S+)\s*$')
    create_branch_re = re.compile('^create-branch (\S+/\S+)\s+(\S+/\S+)\s*$')
    create_version_re = re.compile('^create-version (\S+/\S+)\s*$')
    patch_re = re.compile('^patch (\S+/\S+)\s*$')

    def asHTML(self):
        """Return an HTML representation of the command."""
        result = ''
        for line in self.lines:
            result += cgi.escape(line)
        return result

    def __eq__(self, other):
        return (super(CommandRunner, self).__eq__(other) and
                isinstance(other, CommandRunner) and
                self.lines == other.lines)

    def __init__(self, script, branch_spec_handler, configp, lines):
        """Construct a CommandRunner.

        :param spec_handler: A BranchSpecOptionHandler setting policy for this
                             CommandRunner.
        :param lines: The lines to run.
        """
        super(CommandRunner, self).__init__(script, branch_spec_handler, configp)
        self.lines = lines
        self.arch_impl = None
        self.arch_path = 'baz'
        self._vcs = None
        self.successful = []

    def run(self):
        super(CommandRunner, self).run()
        self.cleanup_wd()
        self.accumulating_patch = False
        self.patch_target = None
        self.patch_content = []
        for line in self.lines:
            self.run_command(line)
        if self.patch_content != []:
            start_time = datetime.now()
            self.patch_target, config = self.getBranchConfig(self.patch_target)
            self.set_current_vcs(self.patch_target)
            self.validate_revision(self.patch_target)
            self.script.logger.info("getting working dir for %s", self.patch_target)
            origdir = os.getcwd()
            dir = self.get_wd(self.script.getSender(),
                              self.patch_target,
                              config)
            try:
                os.chdir(dir)
                try:
                    self.output += ['\n']
                    (summary, moreoutput) = self.do_patch(self.script.getSender(),
                                                          self.patch_content)
                    self.output += moreoutput
                    self.output += ['\n']
                except PQMTlaFailure, e:
                    raise PQMCmdFailure(self.script.getSender(),
                                        self.successful,
                                        'patch ' + self.patch_target,
                                        self.output + e.output)
            except:
                os.chdir(origdir)
                raise
            self.script.logger.info("executing patch")
            self.run_precommit(self.patch_target, config, line, dir)
            self.successful.append('patch ' + self.patch_target)
            self.output += self._success_output('patch', start_time)
            os.chdir(origdir)
            self.get_vcs().commit(self.script.getSender(), dir, summary,
                                  self.patch_target, config)
        self.cleanup_wd()
        return CommandResult(self.successful, self.unrecognized, self.output, not self.reported)

    def run_command(self, line):
        # its a command of some sort
        start_time = datetime.now()
        patch_match = self.patch_re.match(line)
        replay_match = self.replay_re.match(line)
        repo_cache_match=self.repo_cache_re.match(line)
        repo_uncache_match=self.repo_uncache_re.match(line)
        tag_match=self.tag_re.match(line)
        create_branch_match=self.create_branch_re.match(line)
        make_repo_match=self.make_repo_re.match(line)
        create_version_match=self.create_version_re.match(line)
        logger = self.script.logger

        if patch_match:
            # GNU Patch
            logger.info("patch content found, target: %s", patch_match.group(1))
            self.patch_target = patch_match.group(1)
            self.check_target(self.patch_target, line)
            self.accumulating_patch = True
        elif self.accumulating_patch:
            self.patch_content.append(line)
        elif replay_match:
            self.do_merge(from_repo_revision=replay_match.group(1),
                          to_repo_revision=replay_match.group(2),
                          merge_name='replay',
                          merge_method="do_replay",
                          line=line)
        elif repo_cache_match:
            # Cache a revision
            repo_revision = repo_cache_match.group(1)
            (repo, revision) = repo_revision.split('/', 1)
            self.check_target(repo_revision, line)
            self.set_current_vcs(repo_revision)
            self.validate_revision(repo)
            self.output += ['\n', 'Executing repo-cache-revision %s/%s at %s' % (repo, revision,
                                                                               time.strftime('%c')), '\n']
            self.wrap_command(self.get_vcs().do_repo_cache, line, self.sender, repo, revision)
            logger.info("success: %s" % (line,))
            self.successful.append(line)
            self.output += self._success_output('repo-cache-revision', start_time)
        elif repo_uncache_match:
            # Uncache a revision
            repo_revision = repo_uncache_match.group(1)
            (repo, revision) = repo_revision.split('/', 1)
            self.check_target(repo_revision, line)
            self.set_current_vcs(repo_revision)
            self.validate_revision(repo)
            self.output += ['\n', 'Executing repo-uncache-revision %s/%s at %s' % (repo, revision,
                                                                                 time.strftime('%c')), '\n']
            self.wrap_command(self.get_vcs().do_repo_uncache, line, self.sender, repo, revision)
            logger.info("success: %s" % (line,))
            self.successful.append(line)
            self.output += self._success_output('repo-uncache-revision', start_time)
        elif tag_match:
            # Tag a branch
            from_repo_revision = tag_match.group(1)
            to_repo_revision = tag_match.group(2)
            (from_repo, from_revision) = from_repo_revision.split('/', 1)
            (to_repo, to_revision) = to_repo_revision.split('/', 1)
            self.set_current_vcs(from_repo_revision, to_repo_revision)
            self.validate_revision(from_repo_revision)
            self.validate_revision(to_repo_revision)
            self.check_target(to_repo_revision, line)
            self.output += ['\n', 'Executing tag %s/%s at %s' % (from_repo, from_revision,
                                                            time.strftime('%c')), '\n']
            self.wrap_command(self.get_vcs().do_tag, line, self.sender, from_repo,from_revision, to_repo, to_revision)
            logger.info("success: %s" % (line,))
            self.successful.append(line)
            self.output += self._success_output('tag', start_time)
        elif create_branch_match:
            # Create a branch
            from_repo_revision = create_branch_match.group(1)
            to_repo_revision = create_branch_match.group(2)
            self.set_current_vcs(from_repo_revision, to_repo_revision)
            self.validate_revision(from_repo_revision)
            self.validate_revision(to_repo_revision)
            self.check_target(to_repo_revision, line)
            logger.info("getting working dir for %s", to_repo_revision)
            to_repo_revision, config = self.getBranchConfig(to_repo_revision)
            (to_repo, to_revision) = to_repo_revision.split('/', 1)
            dir = self.get_wd(self.sender, to_repo_revision, config)
            origdir = os.getcwd()
            self.output += ['\n',
                            'Executing create-branch %s %s at %s'
                            % (from_repo_revision,
                               to_repo_revision,
                               time.strftime('%c')),
                            '\n']
            try:
                os.chdir(dir)
                self.wrap_command(self.get_vcs().do_create_branch, line, self.sender, to_repo, to_revision)
            except:
                os.chdir(origdir)
                raise
            os.chdir(origdir)
            logger.info("success: %s" % (line,))
            self.successful.append(line)
            self.output += self._success_output('create-branch', start_time)
            os.chdir(origdir)
            self.get_vcs().commit(self.sender, dir, self.commitmsg, to_repo_revision, config)
        elif make_repo_match:
            # Make a repo
            repo = make_repo_match.group(1)
            location = make_repo_match.group(2)
            # make-repo needs its syntax changed - we can't probe for
            # the vcs as the target doesn't exist yet.
            self.check_target(repo, line)
            self.output += ['\n', 'Executing make-repo %s %s at %s' % (repo, location,
                                                                     time.strftime('%c')), '\n']
            self.wrap_command(self.get_arch_impl().do_make_repo, line, self.sender, repo, location)
            logger.info("success: %s" % (line,))
            self.successful.append(line)
            self.output += self._success_output('make-repo', start_time)
        elif create_version_match:
            # Create a new line of development
            repo_revision = create_version_match.group(1)
            (repo, revision) = repo_revision.split('/', 1)
            self.set_current_vcs(repo_revision)
            self.validate_revision(repo_revision)
            self.check_target(repo_revision, line)
            logger.info("getting working dir for %s/%s" % (repo, revision))

            # TODO: Tim Penhey, 2008-05-28
            # Get rid of the imports from pqm module itself.
            from pqm import workdir

            dirpath=os.path.join(workdir, repo)

            if not os.access(dirpath, os.W_OK):
                os.mkdir(dirpath)

            os.chdir(dirpath)
            dir=os.path.join(dirpath, revision)
            if os.access(dir, os.W_OK):
                raise PQMException(self.sender, "Working dir already exists: " + dir)
            os.mkdir(dir)

            origdir = os.getcwd()
            self.output += ['\n', 'Executing create-version %s/%s at %s' % (repo, revision, time.strftime('%c')), '\n']
            try:
                os.chdir(dir)
                self.wrap_command(self.get_arch_impl().do_create_version, line, self.sender, repo, revision)
            except:
                os.chdir(origdir)
                raise
            os.chdir(origdir)
            logger.info("success: %s" % (line,))
            self.successful.append(line)
            self.output += self._success_output('create-version', start_time)
        else:
            self.unrecognized.append(line)

    def write_lines_to_fd(self, lines):
        (stdin, stdout) = os.pipe()
        pid = os.fork()
        if pid != 0:
            os.close(stdout)
            return stdin
        os.close(stdin)
        for line in lines:
            os.write(stdout, line)
            os.write(stdout, '\n')
        os._exit(0)

    def do_patch(self, sender, content):
        # TODO: Tim Penhey, 2008-05-28
        # Get rid of the imports from pqm module itself.
        from pqm import gnupatch_path, popen_noshell_with_input

        def is_patchline(line):
            return line != '' and (line[0] in ('+', '-') or line[0:2] == '@@')
        if content == []:
            raise PQMException(self.sender, "Empty patch content")
        if not is_patchline(content[0]):
            summary = content[0]
        else:
            raise PQMException(self.sender, "No summary given for patch")
        filenames = []
        for line in content:
            if line[0:4] in ('+++ ', '--- '):
                # We intentionally include the date, etc - stripping it out is too hard and error-prone
                filenames.insert(0, line[4:].strip())
        for filename in filenames:
            if (filename.find('/..') > 0) or (filename.find('../') > 0):
                raise PQMException(self.sender, "Invalid backreferencing filename in patch: %s", filename)
            elif filename[0] == '/':
                raise PQMException(self.sender, "Invalid absolute filename in patch: %s", filename)
        fd = self.write_lines_to_fd(content)
        (status, msg, output) = popen_noshell_with_input(gnupatch_path, fd, '-p1', '--batch', '--no-backup-if-mismatch')
        os.close(fd)
        if not ((status is None) or (status == 0)):
            raise PQMException(self.sender, ["patch command \"%s\" failed (%s): %s" % (gnupatch_path, status, msg)] + output)
        return (summary, output)
