# -*- mode: python; coding: utf-8 -*-
# vim:smartindent cinwords=if,elif,else,for,while,try,except,finally,def,class:ts=4:sts=4:sta:et:ai:shiftwidth=4
#
# Copyright © 2008 Canonical Ltd.
#	Author: Tim Penhey <tim@penhey.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging
import os

from twisted.trial import unittest

from pqm.commandline import (
    DEFAULT_CONFIG_FILES,
    normalise_filename,
    parse_command_line,
    )


class TestCommandLineVerbosity(unittest.TestCase):

    def test_defaults(self):
        (options, args) = parse_command_line([])
        self.assertEqual([], args)
        self.assertEqual(logging.WARNING, options.loglevel)

    def test_quiet(self):
        (options, args) = parse_command_line(['-q'])
        self.assertEqual(logging.ERROR, options.loglevel)
        (options, args) = parse_command_line(['--quiet'])
        self.assertEqual(logging.ERROR, options.loglevel)

    def test_very_quiet(self):
        (options, args) = parse_command_line(['-qq'])
        self.assertEqual(logging.CRITICAL, options.loglevel)
        (options, args) = parse_command_line(['-q', '-q'])
        self.assertEqual(logging.CRITICAL, options.loglevel)
        (options, args) = parse_command_line(['--quiet', '--quiet'])
        self.assertEqual(logging.CRITICAL, options.loglevel)

    def test_too_quiet(self):
        (options, args) = parse_command_line(['-qqqq'])
        self.assertEqual(logging.CRITICAL, options.loglevel)
        (options, args) = parse_command_line(['-q', '-q', '-q', '-q'])
        self.assertEqual(logging.CRITICAL, options.loglevel)
        (options, args) = parse_command_line(['--quiet', '--quiet', '--quiet'])
        self.assertEqual(logging.CRITICAL, options.loglevel)

    def test_verbose(self):
        (options, args) = parse_command_line(['-v'])
        self.assertEqual(logging.INFO, options.loglevel)
        (options, args) = parse_command_line(['--verbose'])
        self.assertEqual(logging.INFO, options.loglevel)

    def test_very_verbose(self):
        (options, args) = parse_command_line(['-vv'])
        self.assertEqual(logging.DEBUG, options.loglevel)
        (options, args) = parse_command_line(['-v', '-v'])
        self.assertEqual(logging.DEBUG, options.loglevel)
        (options, args) = parse_command_line(['--verbose', '--verbose'])
        self.assertEqual(logging.DEBUG, options.loglevel)

    def test_too_verbose(self):
        (options, args) = parse_command_line(['-vvvv'])
        self.assertEqual(logging.DEBUG, options.loglevel)
        (options, args) = parse_command_line(['-v', '-v', '-v', '-v'])
        self.assertEqual(logging.DEBUG, options.loglevel)
        (options, args) = parse_command_line(['--verbose', '--verbose', '--verbose'])
        self.assertEqual(logging.DEBUG, options.loglevel)

    def test_mixed_verbosity(self):
        """Handled, but not likely that the user will do this."""
        (options, args) = parse_command_line(['-qv'])
        self.assertEqual(logging.WARNING, options.loglevel)
        (options, args) = parse_command_line(['-qvv'])
        self.assertEqual(logging.INFO, options.loglevel)
        (options, args) = parse_command_line(['-qqv'])
        self.assertEqual(logging.ERROR, options.loglevel)


class TestCommandLineConfigFiles(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        # Move the environment out of the way.
        self.old_environ = os.environ
        os.environ = {}

    def tearDown(self):
        # Restore system environment.
        os.environ = self.old_environ
        unittest.TestCase.tearDown(self)

    def test_defaults(self):
        (options, args) = parse_command_line([])
        self.assertEqual(sorted([normalise_filename(filename)
                                 for filename in DEFAULT_CONFIG_FILES]),
                         sorted(options.configfile_names))

    def test_environment_default(self):
        os.environ['PQM_CONFIG'] = '/usr/src/omg.conf'
        (options, args) = parse_command_line([])
        self.assertEqual(['/usr/src/omg.conf'], options.configfile_names)

    def test_one_config_file(self):
        (options, args) = parse_command_line(['-c', '/usr/src/omg.conf'])
        self.assertEqual(['/usr/src/omg.conf'], options.configfile_names)
        (options, args) = parse_command_line(['--config=/usr/src/omg.conf'])
        self.assertEqual(['/usr/src/omg.conf'], options.configfile_names)

    def test_config_filename_expansion(self):
        (options, args) = parse_command_line(['--config=~/foo/../omg.conf'])
        self.assertEqual([normalise_filename('~/omg.conf')],
                          options.configfile_names)

    def test_two_config_files(self):
        (options, args) = parse_command_line(
            ['-c', '/usr/src/omg.conf', '-c', '/usr/src/wtf.conf'])
        self.assertEqual(['/usr/src/omg.conf', '/usr/src/wtf.conf'],
                         options.configfile_names)
