import logging
import os
import shutil
from StringIO import StringIO
from textwrap import dedent

from bzrlib.tests import TestCaseWithTransport
import config_manager
from twisted.trial import unittest

import pqm
from pqm.errors import PQMCmdFailure, PQMException
from pqm.PQMConfigParser import ConfigParser
from pqm.script import (
    Command, CommandRunner, DebugCommand, EmailScript, MergeCommand)


sample_message = dedent("""\
    From: John.Citizen@example.com
    Subject: A new action
    star-merge http://www.example.com/foo/bar http://www.example.com/bar/baz
    """)
sample_message_2 = dedent("""\
    From: John.Citizen@example.com
    Subject: A new action
    star-merge http://www.example.com/qux/thud http://www.example.com/fred/waldo
    """)
sample_message_bad = dedent("""\
    From: John.Citizen@example.com
    Subject: A new action
    star-merge http://www.example.com/foo/bar http://www.example.com/foo/unregistered
    """)
sample_signed_message = dedent("""\
    From: whee@bar.com (Matthew Thomas)
    Subject: [trivial] fix various actions portlet icons
    -----BEGIN PGP SIGNED MESSAGE-----
    Hash: SHA1

    star-merge http://www.example.com/argh/blah http://www.example.com/bing/bong
    -----BEGIN PGP SIGNATURE-----
    Version: GnuPG v1.2.5 (GNU/Linux)

    iD8DBQFC8guw6PUxNfU6ecoRAt5OAKDPyd1Ac0H7QO4DIEiZ2wvscqrg0ACeNL0h
    ksM1dJeD/aX54NnE/uqps/M=
    =X7Ib
    -----END PGP SIGNATURE-----
    """)

sample_config_with_overrides = dedent("""\
    [DEFAULT]
    queuedir=/tmp
    [location overrides]
    /local/path=sftp://host/otherpath
    scheme://something=righthandside
    """)

sample_config_default_project = dedent("""\
    [DEFAULT]
    queuedir=/tmp
    project=default-project
    [/dummy/]
    """)

sample_config_two_project_specs = dedent("""\
    [DEFAULT]
    queuedir=/tmp
    [/one/]
    project=project-one
    [/two/]
    project=project-two
    [/three/]
    project=project-two
    """)

sample_config_default_and_spec_projects = dedent("""\
    [DEFAULT]
    queuedir=/tmp
    project=default-project
    [/one/]
    project=project-one
    [/two/]
    project=project-two
    [/three/]
    project=project-two
    """)

sample_config_unpublished = dedent("""\
    [DEFAULT]
    queuedir=/tmp
    [http://www.example.com/bar/baz]
    """)

sample_config_published = dedent("""\
    [DEFAULT]
    queuedir=/tmp
    [http://www.example.com/bar/baz]
    published_at=file:///a
    [file:///tmp/foo]
    published_at=file:///b
    publish_to=sftp://host/b
    [file:///tmp/bar/]
    published_at=file:///c/
    """)

sample_config_with_groups = dedent("""\
    [DEFAULT]
    queuedir=/tmp
    groups=agroup
    [agroup]
    members=foo@bar.com
    """)

sample_config_with_workdir = dedent("""\
    [DEFAULT]
    queuedir=queue
    workdir=test-workdir
    """)


class QueueSetup(object):
    """Setup a queue with mock messages in it."""

    def __init__(self):
        self.configFileName = "Foo"
        self.cwd = os.path.abspath(os.path.curdir)
        self.message = sample_message
        self.message_bad = sample_message_bad
        self.message3 = sample_message_2

    def setUp(self):
        myFile=open(self.configFileName, "w")
        myFile.write(dedent("""\
            [DEFAULT]
            queuedir=%s/queue
            [http://www.example.com/bar/baz]
            project=project
            [http://www.example.com/fred/waldo]
            """) % self.cwd)
        myFile.close()
        self.queuedir = os.path.join(self.cwd, "queue")
        try:
            os.mkdir(self.queuedir)
            os.mkdir(self.queuedir + '/pqm')
        except OSError:
            pass
        self.messageFileName = os.path.join(self.queuedir, "patch.00000001")
        messageFile = open(self.messageFileName, "w")
        messageFile.write(self.message)
        messageFile.close()
        self.messageFileNameBad = os.path.join(self.queuedir, "patch.00000002")
        messageFile = open(self.messageFileNameBad, "w")
        messageFile.write(self.message_bad)
        messageFile.close()
        self.messageFileName3 = os.path.join(self.queuedir, "patch.00000003")
        messageFile = open(self.messageFileName3, "w")
        messageFile.write(self.message3)
        messageFile.close()

    def tearDown(self):
        os.unlink(self.configFileName)
        shutil.rmtree(self.queuedir, ignore_errors=True)


class TestWithQueueDirectory(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.queue = QueueSetup()
        self.queue.setUp()

    def testName(self):
        patch = EmailScript('foo.script', logging, False, 0, None, None)
        self.assertEqual(patch.filename, 'foo.script')
        self.scriptname = 'fpp'

    def tearDown(self):
        try:
            os.unlink(self.scriptname)
        except OSError:
            pass
        self.queue.tearDown()

    def getScript(self, content):
        """Get a script for testing with."""
        self.scriptname = 'foo.script'
        scriptFile = open(self.scriptname, "w")
        scriptFile.write(content)
        scriptFile.close()
        pqm.pqm_subdir = os.path.abspath('queue/pqm')
        configp = ConfigParser()
        configp.read([self.queue.configFileName])
        handler = pqm.BranchSpecOptionHandler(configp)
        return EmailScript(self.scriptname, logging, False, 54, handler, configp)

    def testFields(self):
        script = self.getScript(sample_message)
        self.assertEqual(script.getSender(), "John.Citizen@example.com")
        self.assertEqual(script.getSubject(), "A new action")
        self.assertEqual(script.getContent(),
            ("star-merge http://www.example.com/foo/bar "
             "http://www.example.com/bar/baz\n"))
        self.assertEqual(script.getLines(),
            [("star-merge http://www.example.com/foo/bar "
              "http://www.example.com/bar/baz")])
        self.assertEqual([MergeCommand(None,
                                       None,
                                       None,
                                       'http://www.example.com/foo/bar',
                                       'http://www.example.com/bar/baz')],
                         script.getCommands())

    def testGPGFields(self):
        script = self.getScript(sample_signed_message)
        self.assertEqual(script.getSender(), "whee@bar.com (Matthew Thomas)")
        self.assertEqual(script.getSubject(),
            "[trivial] fix various actions portlet icons")
        self.assertEqual(script.getContent(), dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            star-merge http://www.example.com/argh/blah http://www.example.com/bing/bong
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1.2.5 (GNU/Linux)

            iD8DBQFC8guw6PUxNfU6ecoRAt5OAKDPyd1Ac0H7QO4DIEiZ2wvscqrg0ACeNL0h
            ksM1dJeD/aX54NnE/uqps/M=
            =X7Ib
            -----END PGP SIGNATURE-----
            """))
        self.assertEqual(script.getLines(), 
            [("star-merge http://www.example.com/argh/blah "
              "http://www.example.com/bing/bong")])
        self.assertEqual(
            [MergeCommand(None,
                          None,
                          None,
                          'http://www.example.com/argh/blah',
                          'http://www.example.com/bing/bong')],
            script.getCommands())

    def testDate(self):
        """Can we access a submission time for scripts."""
        script = self.getScript(sample_message)
        self.assertEqual(script.getSubmissionTime(), 54)

    def testProjects(self):
        # A script may affect multiple branches, so we can ask the script
        # for its projects, and then for the commands for a project.
        script = self.getScript(sample_message)
        self.assertEqual(set(['project']), script.getProjects())

    def test_log_status(self):
        # make an initial status file
        status = file(self.queue.queuedir + '/pqm/status', 'wt')
        status.write('bar\n')
        status.close()
        script = self.getScript(sample_message)
        command = Command(script,
                          script._branch_spec_handler,
                          script._configp)
        logger = MockLogger()
        command.log_with_status(logger, 'foo')
        content = file(self.queue.queuedir + '/pqm/status').read()
        self.assertEqual('bar\nfoo\n', content)



# NOTDONEYET: move command recognition from CommandRunner to Script
# and commands should follow the command pattern rather than being strings.
#
#    def testIsCommand(self):
#        """Is a line a command line?"""
#        import pqm
#        script = pqm.Script('foo.script', logging, False)
#        self.failIf(script.is_command(""))


class MockLogger(object):

    def __init__(self):
        self.logs = []

    def info(self, message, *args, **kwargs):
        """Pretend to work."""
        self.logs.append(message)


class MockScript(object):

    def __init__(self):
        """Create a Script which is only useful for testing."""
        self.debug = False
        self.logger = MockLogger()
        self.signing_email = 'gpg_sender'

    def getSubject(self):
        return "a subject."

    def getSender(self):
        return "a_sender"


class TestCommandRunner(unittest.TestCase):

    def test_star_merge_urls(self):
        star_match = EmailScript.star_re.match("star-merge file:///url1 file:///url2")
        self.assertEqual(star_match.group(1), 'file:///url1')
        self.assertEqual(star_match.group(2), 'file:///url2')

    def test_get_arch_impl(self):
        self.logs = []
        configp = ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        script = MockScript()
        runner = CommandRunner(script, handler, configp, [])
        # TODO should be able to pass on systems without baz installed.
        try:
            self.failUnless(isinstance(runner.get_arch_impl(),
                                       pqm.Baz1_1Handler))
        except PQMException, e:
            if "Could not determine arch" not in str(e):
                # something went wrong other than baz not being installed.
                raise


    def test_check_revision(self):
        configp = ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        handler._specs = {'file:///tmp/foo':{},
                          'file:///tmp/bar/':{}}
        runner = CommandRunner(None, handler, configp, [])
        runner.script = MockScript()
        runner.check_target('file:///tmp/foo', 'blah')
        self.assertRaises(PQMCmdFailure,
                          runner.check_target,
                          'file:///tmp/foo/branch',
                          'blah')
        self.assertRaises(PQMCmdFailure,
                          runner.check_target,
                          'file:///tmp/bar',
                          'blah')
        runner.check_target('file:///tmp/bar/branch', 'blah')

    def test__make_wd_path(self):
        configp = ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(None, handler, configp, [])
        self.assertEqual(runner._make_wd_path('/foo', '/foo/bar'), '/foo/foo')
        self.assertEqual(runner._make_wd_path('/foo', 'file:///foo/bar'),
                         '/foo/file')
        self.assertEqual(runner._make_wd_path('/foo', 'branch'), '/foo/branch')

    def test__branch_name(self):
        configp = ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(None, handler, configp, [])
        self.assertEqual(runner._branch_name('foo'), 'foo')
        self.assertEqual(runner._branch_name('file:///home/bar/foo'), 'foo')
        self.assertEqual(runner._branch_name('file:///home/bar/foo'), 'foo')

    def test_cleanup_wd_missing_dir(self):
        configp = ConfigParser()
        configp.readfp(StringIO(sample_config_with_workdir))
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(MockScript(), handler, configp, [])
        pqm.workdir = "test-workdir"
        pqm.branch_specs={'path':{'build_dir':'nonexistant-dir'}}
        os.mkdir("test-workdir")
        try:
            runner.cleanup_wd()
        finally:
            shutil.rmtree("test-workdir")

    def test__get_url_override_mapper_no_overrides(self):
        configp = ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(None, handler, configp, [])
        runner.configp = ConfigParser()
        expected_mapper = config_manager.URLMapper()
        self.assertEqual(expected_mapper, runner._get_url_override_mapper())

    def test__get_url_override_mapper(self):
        configp = ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(None, handler, configp, [])
        runner.configp = ConfigParser()
        runner.configp.readfp(StringIO(sample_config_with_overrides))
        expected_mapper = config_manager.URLMapper()
        expected_mapper.add_map("sftp://host/otherpath", "/local/path")
        expected_mapper.add_map("righthandside", "scheme://something")
        self.assertEqual(expected_mapper, runner._get_url_override_mapper())


class TestCommand(unittest.TestCase):

    def test_debug_command_sets_script_debug(self):
        configp = ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        script = MockScript()
        self.assertFalse(script.debug)
        command = DebugCommand(script, handler, configp)
        result = command.run()
        self.assertEqual(['debug'], result.successes)
        self.assertEqual([], result.unrecognized)
        self.assertEqual(['Debug mode enabled.'], result.output)
        self.assertTrue(result.report)
        self.assertTrue(script.debug)


class FunctionalTestCommandRunner(unittest.TestCase):

    def setUp(self):
        from bzrlib.bzrdir import BzrDir
        from bzrlib.plugin import load_plugins
        load_plugins()
        tree = BzrDir.create_standalone_workingtree("bzrbranch")
        tree.smart_add(["bzrbranch"])
        tree.commit("start branch.", verbose=False)

    def tearDown(self):
        shutil.rmtree("bzrbranch")

    def test_get_branch_handler_arch(self):
        # I would test this, but I don't want to add a full dependency on
        # pybaz at this point
        #from pqm import CommandRunner
        #runner = CommandRunner()
        ## TODO set a fake logger
        pass

    def branch_url(self):
        branchurl = "file://%s" % os.path.abspath("bzrbranch")
        return branchurl.replace('\\', '/')

    def test_get_branch_handler_bzr_file(self):
        configp = ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(None, handler, configp, [])
        runner.script = MockScript()
        self.failUnless(isinstance(runner.get_branch_handler(self.branch_url()),
                                   pqm.Bazaar2Handler))

    def test_get_branch_handler_bzr_relative(self):
        configp = ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(None, handler, configp, [])
        runner.script = MockScript()
        self.failUnless(isinstance(runner.get_branch_handler("bzrbranch"),
                                   pqm.Bazaar2Handler))

    def test_set_current_vcs(self):
        configp = ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(None, handler, configp, [])
        runner.script = MockScript()
        self.assertEqual(runner.get_vcs(), None)
        runner.set_current_vcs(self.branch_url(), "bzrbranch")
        self.assertNotEqual(runner.get_vcs(), None)
        self.failUnless(isinstance(runner.get_vcs(), pqm.Bazaar2Handler))


class BzrHandlerTestCase(TestCaseWithTransport):

    def setUp(self):
        super(BzrHandlerTestCase, self).setUp()
        from bzrlib.bzrdir import BzrDir
        from bzrlib.plugin import load_plugins
        load_plugins()
        tree = self.make_branch_and_tree("bzrbranch")
        tree.smart_add(["bzrbranch"])
        tree.commit("start branch.", verbose=False)
        tree.branch.bzrdir.sprout("bzrbranch-parent")
        tree.branch.bzrdir.sprout("bzrbranch-public")
        tree.branch.bzrdir.sprout("branch-contributor")
        contrib_tree = BzrDir.open("branch-contributor").open_workingtree()
        self.build_tree_contents([("branch-contributor/README", "Boo!\n")])
        contrib_tree.add(["README"])
        contrib_tree.commit("add README")

    def tearDown(self):
        shutil.rmtree("bzrbranch")
        shutil.rmtree("branch-contributor")
        shutil.rmtree("bzrbranch-parent")
        shutil.rmtree("bzrbranch-public")
        super(BzrHandlerTestCase, self).tearDown()

    def test_commit(self):
        from bzrlib.branch import Branch
        branch = Branch.open("bzrbranch")
        self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                         'start branch.')
        handler = pqm.Bazaar2Handler()
        config = {'publish_to':'bzrbranch-public'}
        handler.commit("me", "bzrbranch", "wheee", "bzrbranch-parent", config)
        for path in ['bzrbranch', 'bzrbranch-parent', 'bzrbranch-public']:
            branch = Branch.open(path)
            rev = branch.last_revision()
            self.assertEqual(branch.repository.get_revision(rev).message, 'wheee')

    def test_commit_uncommits_when_push_fails(self):
        from bzrlib.branch import Branch
        branch = Branch.open("bzrbranch")
        self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                         'start branch.')
        handler = pqm.Bazaar2Handler()
        config = {'publish_to':'bzrbranch-public-missing'}
        self.assertRaises(pqm.PQMTlaFailure, handler.commit, "me",
                          "bzrbranch", "wheee", "bzrbranch-parent", config)
        # bzr branch's commit is allowed as its the local tree - we ont
        # care what happens there, but the
        # parent and published copies must be reverted.
        for path in ['bzrbranch-parent', 'bzrbranch-public']:
            branch = Branch.open(path)
            rev = branch.last_revision()
            self.assertEqual(branch.repository.get_revision(rev).message, 'start branch.')

    def test_make_local_dir(self):
        try:
            handler = pqm.Bazaar2Handler()
            handler.make_local_dir("me", "bzrbranch", "proof")
            from bzrlib.branch import Branch
            branch = Branch.open("proof")
            self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                             'start branch.')
        finally:
            shutil.rmtree("proof")

    def test_merge(self):
        from bzrlib.branch import Branch
        branch = Branch.open("bzrbranch")
        self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                         'start branch.')
        contrib_branch = Branch.open("branch-contributor")
        self.assertEqual(contrib_branch.repository.get_revision(
                            contrib_branch.last_revision()).message,
                         'add README')
        self.failIf(os.path.exists("bzrbranch/README"))
        handler = pqm.Bazaar2Handler()
        result = handler.do_star_merge("me", "branch-contributor", "bzrbranch")
        self.assertEqual(result, ["merge successful"])
        branch = Branch.open("bzrbranch")
        self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                         'start branch.')
        contrib_branch = Branch.open("branch-contributor")
        self.assertEqual(contrib_branch.repository.get_revision(
                            contrib_branch.last_revision()).message,
                         'add README')
        self.failUnless(os.path.exists("bzrbranch/README"))

    def test_merge_conflicts(self):
        from bzrlib.workingtree import WorkingTree
        from bzrlib.branch import Branch
        tree = WorkingTree.open('bzrbranch')
        self.assertEqual(tree.branch.repository.get_revision(tree.last_revision()).message,
                         'start branch.')
        contrib_branch = Branch.open("branch-contributor")
        self.assertEqual(contrib_branch.repository.get_revision(
                            contrib_branch.last_revision()).message,
                         'add README')
        contrib_tree = WorkingTree.open('branch-contributor')
        # give tree a README with common ancestry
        tree.pull(contrib_branch)
        # commit new READMEs to both trees.
        file = open("bzrbranch/README", "wt")
        file.write("Boo conflict!\n")
        file.close()
        tree.commit('add conflicting README')

        file = open("branch-contributor/README", "wt")
        file.write("Boo conflict YEAH!\n")
        file.close()
        contrib_tree.commit('add conflicting README')
        handler = pqm.Bazaar2Handler()
        err = self.assertRaises(pqm.PQMTlaFailure, handler.do_star_merge, "me",
                                "branch-contributor", "bzrbranch")
        self.assertEqual(["Conflicts during merge: ",
                          "Text conflict in README",
                          ],
                         err.output)

    def test_merge_unrelated(self):
        unrelated = self.make_branch_and_tree('unrelated')
        unrelated.commit("unrelated first commit")
        handler = pqm.Bazaar2Handler()
        try:
            handler.do_star_merge("me", "unrelated", "bzrbranch")
        except pqm.PQMTlaFailure, e:
            self.assertEqual(
                ["Branches have no common ancestor, and no merge base"
                 " revision was specified."],
                 e.output)
            return
        self.fail("Merge base error not raised.")

    def test_latest_revision_info(self):
        handler = pqm.Bazaar2Handler()
        revno, message = handler.latest_revision_info("me", "bzrbranch")
        self.assertEqual(1, revno)
        self.assertEqual('start branch.', message)


class TestConfig(unittest.TestCase):

    def test_location_overrides(self):
        config = ConfigParser()
        config.readfp(StringIO(sample_config_with_overrides))
        location_rules = config.section_items('location overrides')
        expected_rules = [("/local/path", "sftp://host/otherpath"),
                          ("scheme", "//something=righthandside")]
        self.assertEqual(set(expected_rules), set(location_rules))

    def test_projects_empty(self):
        # empty config -> no projects
        configp = ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        branch_spec_handler = pqm.BranchSpecOptionHandler(configp)
        self.assertEqual(set(), branch_spec_handler.project_names())

    def test_projects_default(self):
        # default project -> its listed and present as a branch option.
        configp = ConfigParser()
        configp.readfp(StringIO(sample_config_default_project))
        pqm.pqm_subdir = '/tmp' # ewww
        branch_spec_handler = pqm.BranchSpecOptionHandler(configp)
        self.assertEqual(set(['default-project']),
                         branch_spec_handler.project_names())
        self.assertEqual('default-project',
            branch_spec_handler.get_target_config('/dummy/foo')[1]['project'])

    def test_projects_branch_specs(self):
        # branch specs have their projects aggregated and dont error with
        # one or two specs mentionining a project.
        configp = ConfigParser()
        configp.readfp(StringIO(sample_config_two_project_specs))
        pqm.pqm_subdir = '/tmp' # ewww
        branch_spec_handler = pqm.BranchSpecOptionHandler(configp)
        self.assertEqual(set(['project-one', 'project-two']),
                         branch_spec_handler.project_names())
        self.assertEqual('project-one',
            branch_spec_handler.get_target_config('/one/foo')[1]['project'])
        self.assertEqual('project-two',
            branch_spec_handler.get_target_config('/two/foo')[1]['project'])
        self.assertEqual('project-two',
            branch_spec_handler.get_target_config('/two/foo')[1]['project'])

    def test_projects_branch_specs_and_default(self):
        # branch specs have their projects aggregated with the default spec.
        configp = ConfigParser()
        configp.readfp(StringIO(sample_config_default_and_spec_projects))
        pqm.pqm_subdir = '/tmp' # ewww
        branch_spec_handler = pqm.BranchSpecOptionHandler(configp)
        self.assertEqual(set(['default-project', 'project-one', 'project-two']),
                         branch_spec_handler.project_names())
        self.assertEqual('project-one',
            branch_spec_handler.get_target_config('/one/foo')[1]['project'])
        self.assertEqual('project-two',
            branch_spec_handler.get_target_config('/two/foo')[1]['project'])
        self.assertEqual('project-two',
            branch_spec_handler.get_target_config('/two/foo')[1]['project'])

    def test_get_target_unpublished_config(self):
        configp = ConfigParser()
        configp.readfp(StringIO(sample_config_unpublished))
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        branch, config = handler.get_target_config('http://www.example.com/bar/baz')
        self.assertEqual('http://www.example.com/bar/baz', branch)
        self.assertEqual(None, config['published_at'])
        self.assertEqual(None, config['publish_to'])
        self.assertRaises(KeyError, handler.get_target_config, 'baz/baz')

    def test_get_target_config(self):
        configp = ConfigParser()
        configp.readfp(StringIO(sample_config_published))
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        branch, config = handler.get_target_config(
                                            'http://www.example.com/bar/baz')
        self.assertEqual('http://www.example.com/bar/baz', branch)
        self.assertEqual('file:///a', config['published_at'])
        self.assertEqual('file:///a', config['publish_to'])
        # close but shouldn't match:
        self.assertRaises(KeyError, handler.get_target_config, 'bar/bar')
        self.assertRaises(KeyError, handler.get_target_config, 'file:///a/bar')
        # note that publish_to is not overriden by the default from published_at.
        branch, config = handler.get_target_config('file:///tmp/foo')
        self.assertEqual('file:///tmp/foo', branch)
        self.assertEqual('file:///b', config['published_at'])
        self.assertEqual('sftp://host/b', config['publish_to'])
        # get published_at -> real location, publish_to still sftp://
        branch, config = handler.get_target_config('file:///b')
        self.assertEqual('file:///tmp/foo', branch)
        self.assertEqual('file:///b', config['published_at'])
        self.assertEqual('sftp://host/b', config['publish_to'])
        # close but shouldn't match:
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///tmp/foo/branch')
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///b/branch')
        branch, config = handler.get_target_config('file:///tmp/bar/branch')
        self.assertEqual('file:///tmp/bar/branch', branch)
        self.assertEqual('file:///c/branch', config['published_at'])
        self.assertEqual('file:///c/branch', config['publish_to'])
        branch, config = handler.get_target_config('file:///c/branch')
        self.assertEqual('file:///tmp/bar/branch', branch)
        self.assertEqual('file:///c/branch', config['published_at'])
        self.assertEqual('file:///c/branch', config['publish_to'])
        # but the location isn't a branch itself
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///tmp/bar')
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///c')
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///c/')

    def test_smoking_groups(self):
        configp = ConfigParser()
        configp.readfp(StringIO(sample_config_with_groups))
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)


class TestStatusFile(unittest.TestCase):

    def setUp(self):
        super(TestStatusFile, self).setUp()
        if os.path.exists('statusfile'):
            os.unlink('statusfile')

    def test_make_new(self):
        status = pqm.StatusFile('statusfile')
        self.assertTrue(os.path.exists('statusfile'))

    def test_line_has_newline(self):
        #status.line always writes a newline
        status = pqm.StatusFile('statusfile')
        status.line('foo')
        content = file('statusfile').read()
        self.assertEqual('foo\n', content)

    def test_truncates_at_12_lines(self):
        statusfile = file('statusfile', 'wt')
        statusfile.write('long line\n')
        statusfile.write('long line\n')
        statusfile.write('long line\n')
        statusfile.write('long line\n')
        statusfile.write('long line\n')
        statusfile.close()
        # after 5 lines the status shows only the end 5 lines
        # even when they are shorter than the previous content.
        status = pqm.StatusFile('statusfile')
        for line in range(13):
            status.line(str(line))
        content = file('statusfile').read()
        self.assertEqual(
            '1\n'
            '2\n'
            '3\n'
            '4\n'
            '5\n'
            '6\n'
            '7\n'
            '8\n'
            '9\n'
            '10\n'
            '11\n'
            '12\n',
            content)

    def test_open_does_not_truncate(self):
        statusfile = file('statusfile', 'wt')
        statusfile.write('foo\n')
        statusfile.close()
        status = pqm.StatusFile('statusfile')
        content = file('statusfile').read()
        self.assertEqual('foo\n', content)

    def test_set_status_area_fills_height(self):
        statusfile = file('statusfile', 'wt')
        status = pqm.StatusFile('statusfile', 5)
        status.line('1')
        status.set_status_area_height(2)
        content = file('statusfile').read()
        self.assertEqual('\n\n1\n\n\n', content)

    def test_after_set_status_area_height_bottom_N_do_not_scroll(self):
        statusfile = file('statusfile', 'wt')
        status = pqm.StatusFile('statusfile', 5)
        for line in range(3):
            status.line(str(line))
        status.set_status_area_height(2)
        for line in range(3,5):
            status.status_line(str(line))
        status.line('5')
        content = file('statusfile').read()
        self.assertEqual('1\n2\n5\n3\n4\n', content)

    def test_status_line_scrolls_status_area(self):
        statusfile = file('statusfile', 'wt')
        status = pqm.StatusFile('statusfile', 5)
        for line in range(5):
            status.line(str(line))
        status.set_status_area_height(2)
        status.status_line('5')
        status.status_line('6')
        status.status_line('7')
        content = file('statusfile').read()
        self.assertEqual('2\n3\n4\n6\n7\n', content)


class TestWriteToLineAdapter(unittest.TestCase):

    def setUp(self):
        super(TestWriteToLineAdapter, self).setUp()
        self.lines = []
        self.adapter = pqm.WriteToLineAdapter(self)

    def line(self, line):
        self.lines.append(line)

    def test_normal_line_is_line(self):
        self.adapter.write('foo bar\n')
        self.assertEqual(['foo bar\n'], self.lines)

    def teste_unterminated_line_gets_terminated(self):
        self.adapter.write('hiya')
        self.assertEqual(['hiya\n'], self.lines)
