
from twisted.trial import unittest
from twisted.web.resource import getChildForRequest
from twisted.web.test.test_web import DummyRequest
import logging

import pqm
from pqm.PQMConfigParser import ConfigParser
from pqm.core import get_script_queue
from pqm.tests.test_pqm import QueueSetup

class TestTwistedUI(unittest.TestCase):

    def testImports(self):
        import pqm.ui.twistd

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.queueSetup = QueueSetup()
        self.queueSetup.setUp()

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.queueSetup.tearDown()

    def testQueue(self):
        from pqm.ui.twistd import FakeOptions
        configp = ConfigParser()
        configp.read([self.queueSetup.configFileName])
        handler = pqm.BranchSpecOptionHandler(configp)
        queuedir = pqm.get_queuedir(configp, logging, [])

        script_queue = get_script_queue(
            queuedir, logging, handler, configp, FakeOptions())

        patches = script_queue.items()
        self.assertEqual(3, len(patches))
        self.assertEqual(patches[0].filename, self.queueSetup.messageFileName)
        self.assertEqual(set(['project']),
                         patches[0].getCommands()[0].getProjects())
        self.assertRaises(KeyError,
                          patches[1].getCommands()[0].getProjects)
        self.assertEqual(set([]),
                         patches[2].getCommands()[0].getProjects())

    def getResource(self):
        from pqm.ui.twistd import PQMInfo, QueueResource
        statusfile = file(self.queueSetup.queuedir + '/pqm/status', 'wt')
        statusfile.write("<foo\nbar\nbaz\nquux\ntheta\n")
        statusfile.close()
        pqminfo = PQMInfo([self.queueSetup.configFileName])
        resource = QueueResource(pqminfo)
        pqminfo.refresh()
        return resource

    def assertDocTest(self, expected, actual):
        """Run a doctest output comparison."""
        from doctest import OutputChecker, ELLIPSIS, Example
        checker = OutputChecker()
        if not checker.check_output(expected, actual, ELLIPSIS):
            self.fail('Doctest comparison failed: %s' % 
                      checker.output_difference(Example('', expected), actual, ELLIPSIS))

    def testSlash(self):
        # This test tests the appearance of the / web page in the web ui.
        # three scripts are shown:
        # the active script:
        # - A script that belongs to a project should not appear in the
        # / web page, rather a placeholder of 'X requests for other projects'
        # and no output pane is shown.
        # should be shown.
        # - A script that is not for a PQM managed branch should show as an
        #   error.
        # - A script for no projects should be shown normally.
        queue = self.getResource()
        request = DummyRequest([''])
        result_resource = getChildForRequest(queue, request)
        html = result_resource.render(request)
        expected = (
"<h1>PQM Queue: 3 scripts</h1>"
"<p>Current time: ...</p>"
"<h2>Now playing...</h2>"
"<ul>"
"<li><p>1 request(s) for other project(s).</p></li>"
"</ul>"
"<h2>Coming up</h2>"
"<ol>"
"<li><p>...: John.Citizen@example.com, Request for non-PQM managed branch.</p></li>"
"<li><p>...: John.Citizen@example.com, 'A new action' "
"<ol><li>Merge http://www.example.com/qux/thud http://www.example.com/fred/waldo</li></ol></p></li>"
"</ol>")
        self.assertDocTest(expected, html)

    def testProjectSpecificPage(self):
        # This test tests the appearance of the /<projectname> web page in the
        # web ui.
        # three scripts are shown:
        # the active script:
        # - A script that belongs to the project - it should appear normally.
        # a status pane is shown for it.
        # - A script that is not for a PQM managed branch should show as an  
        #   error.
        # - A script for no projects should be shown as are scripts for other
        #   projects - as a 'requiest for other project'
        queue = self.getResource()
        request = DummyRequest(['project', ''])
        result_resource = getChildForRequest(queue, request)
        html = result_resource.render(request)
        expected = (
"<h1>PQM Queue for project: 3 scripts</h1>"
"<p>Current time: ...</p>"
"<h2>Now playing...</h2>"
"<pre>"
"&lt;foo\n"
"bar\n"
"baz\n"
"quux\n"
"theta\n"
"</pre>"
"<ul>"
"<li><p>...: John.Citizen@example.com, 'A new action' "
"<ol><li>Merge http://www.example.com/foo/bar http://www.example.com/bar/baz</li></ol></p></li>"
"</ul>"
"<h2>Coming up</h2>"
"<ol>"
"<li><p>...: John.Citizen@example.com, Request for non-PQM managed branch.</p></li>"
"<li><p>1 request(s) for other project(s).</p></li>"
"</ol>")
        self.assertDocTest(expected, html)
