# vim:smartindent cinwords=if,elif,else,for,while,try,except,finally,def,class:ts=4:sts=4:sta:et:ai:shiftwidth=4
# Copyright (C) 2005 Canonical Limited
#       Authors: Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import cgi
import logging
import os
import sys
import time

from twisted.application import internet, service
from twisted.web import resource, server, static

import pqm
from pqm.PQMConfigParser import ConfigParser
from pqm.commandline import default_pqm_config_files
from pqm.core import get_script_queue

class QueueResource(resource.Resource):
    """A resource that shows a PQM queue."""

    isLeaf = True

    def __init__(self, queue):
        self.queue = queue
        resource.Resource.__init__(self)

    def render(self, request):
        """Get a child of the queue"""
        if len(request.postpath) > 3:
            request.prepath.extend(request.postpath)
            request.postpath[:] = []
            return static.Data("<h1>Please visit the root page or a project "
                               "specific page.</h1>", 'text/html').render(request)
        if len(request.postpath) > 1:
            project = request.postpath[0]
            request.prepath.extend(request.postpath)
            request.postpath[:] = []
            return self.getProjectPage(project, request).render(request)
        else:
            return self.getProjectPage(None, request).render(request)

    def getProjectPage(self, selected_project, request):
        """Render the project page."""
        if selected_project:
            project_text = ' for %s' % selected_project
        else:
            project_text = ''
        text = "<h1>PQM Queue%s: %d scripts</h1>" % (project_text,
            len(self.queue.messages))
        if not self.queue.is_processing_requests:
            text += "<h2>PQM is not currently processing additional requests</h2>"
        text += "<p>Current time: %s UTC</p>" % cgi.escape(
            time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime()))
        other_project_request_count = 0
        if self.queue.messages:
            text += (
                "<h2>Now playing...</h2>"
                )
            more_text, other_project_request_count = self.formatOneScript(
                self.queue.messages[0],
                other_project_request_count,
                selected_project)
            # show the status pane if this project was allowed:
            if not other_project_request_count:
                text += "<pre>"
                text += cgi.escape(self.queue.status_text)
                text += "</pre>"
            text += "<ul>"
            text += more_text
            if other_project_request_count:
                text += self.formatOtherProjectRequests(other_project_request_count)
                other_project_request_count = 0

            text += "</ul>"
        if len(self.queue.messages) < 2:
            return static.Data(text, 'text/html')
        # print a queue of pending messages
        text += (
            "<h2>Coming up</h2>"
            "<ol>"
            )
        for message in self.queue.messages[1:]:
            more_text, other_project_request_count = self.formatOneScript(
                message,
                other_project_request_count,
                selected_project)
            text += more_text
        if other_project_request_count:
            text += self.formatOtherProjectRequests(other_project_request_count)
        text += "</ol>"
        return static.Data(text, 'text/html')

    def formatOneScript(self, message, other_project_request_count, selected_project):
        """Format a single script.
        
        :return: the formatted text, the current accrued other_project_request
                 count
        """
        result = ''
        try:
            projects = message.getProjects()
        except KeyError:
            result += self.formatMessagePrelude(message, other_project_request_count)
            result += "Request for non-PQM managed branch."
        else:
            if ((selected_project is None and len(projects)) or 
                (selected_project is not None and 
                 selected_project not in projects)):
                return '', other_project_request_count + 1
            result += self.formatMessagePrelude(message, other_project_request_count)
            message.getProjects()
            result += "'%s' <ol>" % (
                cgi.escape(message.getSubject()))
            for command in message.getCommands():
                result += "<li>%s</li>" % command.asHTML()
            result += "</ol>"
        result += "</p></li>" 
        return result, 0

    def formatMessagePrelude(self, message, other_project_request_count):
        text = ''
        if other_project_request_count:
            text += self.formatOtherProjectRequests(other_project_request_count)
        text += "<li><p>%s: %s, " % (
            cgi.escape(message.asctime()),
            cgi.escape(message.getSender()))
        return text

    def formatOtherProjectRequests(self, other_project_request_count):
        return '<li><p>%d request(s) for other project(s).</p></li>' % \
            other_project_request_count


class FakeOptions(object):
    verify_sigs = False
    keyring = None


class PQMInfo(object):

    def __init__(self, filenames):
        self.filenames = filenames
        self.messages = []
        self.is_processing_requests = False

    def refresh(self):
        configp = ConfigParser()
        configp.read(self.filenames)
        handler = pqm.BranchSpecOptionHandler(configp)
        queuedir = pqm.get_queuedir(configp, logging, [])

        script_queue = get_script_queue(
            queuedir, logging, handler, configp, FakeOptions())

        self.is_processing_requests = script_queue.is_processing_requests()
        self.messages = script_queue.items()

        try:
            [message.getSender() for message in self.messages]
        except:
            # We dont care about exceptions in getSender() here -
            # we are just preseeding the files to reduce the race
            # condition with script completion.
            pass
        self.status_text = ''.join(handler.status.lines)

application = None
def main(argv):
    global application
    application = service.Application('pqm')
    pqminfo = PQMInfo(default_pqm_config_files())
    serviceCollection = service.IServiceCollection(application)
    port_str = os.environ.get('PQM_PORT', '8000')
    internet.TCPServer(int(port_str), server.Site(QueueResource(pqminfo))
        ).setServiceParent(serviceCollection)
    internet.TimerService(30, pqminfo.refresh).setServiceParent(serviceCollection)

if __name__ == '__builtin__':
    main(sys.argv)
